/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/
/*      peal.h:  Header file for lm_map.c                                 */
/*               Contains functions for peeling and reverse peeling.      */
/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

#define CHAIN 1
#define SMALL 1.0e-06
#define SCALE 1.0e06
#define MXPSQ 650    /* The length of the longest peeling sequence        */
#define BIG 200000   /* As much memory as needed to store peeling fncts   */
#define NP 1200      /* The largest index of all indivs. providing phenos */
#define NG 80        /* The number of possible geno in population         */
#define PNG 6100     /* The number of possible mating types in population */
#define NI 20        /* The largest possible cluster size                 */
#define NST 100      /* The largest number of peeling op. in a peel. seq. */

/* CURRENT CUMULATIVE PENETRANCES, INDIVIDUAL GENOTYPES  */
        double pen[NI+1];
        int igen[NI], invol[NI];

/* CURRENT POSITIONS ETC. OF INDIVIDUALS OF PARENT OFFSPRING TRIPLETS 
   INVOLVED IN CURRENT OPERATION  (# triplets < NI)    */
        int itrip[4][NI], mtrip[NI];
      
/* POSITIONS OF INDIVIDUALS IN R-FUNCTIONS CURRENTLY INVOLVED, AND
   NUMBERS AND POSITIONS IN PRLIK ETC. (# prevlik < NI) */
        int iprlik[NI+1][NI], nprlik[NI], nstlik[NI], maxlik[NI];

/* CHILDEREN IN CURRENTLY INVOLVED SIBSHIP; ONLY CHILDREN PEELED FROM/TO 
        INCLUDED */
        int ich[NI];

/* STORAGE SPACE FOR R-FUNCTIONS, AND OUTPUT R-FUNCTION. 
     FIRST ngenty TERMS OF store CONTAIN R-FUNCTION FOR ANY NOT
     PREVIOUSLY CONSIDERED FOUNDER PARENT. 
   ALSO MARGINAL R-FUNCTIONS FOR INTERMEDIATE NAME SET INDIVIDUALS.  */
        double store[2*BIG], fmarg[NI][NG+1];

/*  NAMES IN STORED R-FUNCTIONS, POSITION IN STORAGE, AND INDEX FOR
      FUTURE REQUIREMENT. */
        int namsto[NST][NI], nstart[NST], ndex[NST], nindi[NST];

/*
  IMPOSS=POSITION OF FIRST INDIVIDUAL FOR WHICH THE CURRENT GENOTYPE
    SET IS IMPOSSIBLE.
  IPLACE=FIRST INDIVIDUAL TO HAVE BEEN RESET IN CURRENT GENOTYPE SET.
  NINVOL= # INDIVIDUALS INVOLVED IN CURRENT OPERATION
  NOUTIN= # INDIVIDUALS IN OUTPUT SET OF CURRENT OPERATION.
  NSTORE = # STORED R-FUNCTIONS (INCLUDING FOUNDER R-FUNCTION)  

      COMMON NGENTY,NSTORE,STORE,NSTART,INDEX,NINDI,NAMSTO   */

   int ngenty, nphen, ngng, npeop, npeph;
   int nstore, nextst, noutin, ntrip, nsib, nlik, npar, ninvol, iwrt;
   int im, ife, iim, iife, ifoum, ifoufe, iplace, iout, imposs;

int code_dna(int, int, int);            /* Converts genos to int codes    */

void uncode_dna(int, int*, int*, int);  /* Convert integer codes to genos */

double sege(int, int, int, int);        /* Computes parent-offspring segs */

double pene(int g, int y, double, int); /* Computes the P(y|g,epsilon)    */

void prev_functions(), repack(), varlik();
int rescale();

double peel_rvsp( int, double, int, int*, int, double*, int***, int* );
                                        /* Peels a locus and/or samples
                                           the genos at that locus        */

