/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/
/*      limdefs.h:  included by stuff.h                                */
/*                  defines MORGAN default values, including minimums  */
/*                      and maximums, for parameters which have to do  */
/*                      with pedigree files, pedigree data and the     */
/*                      the modelling or simulation of this data       */
/*                  many of these parameters may by overriden in       */
/*                      parameter statements                           */
/*                  if any of these values are changed, the MORGAN     */
/*                      libraries and programs should be recompiled    */
/*                      and relinked                                   */
/*                  also includes a few other useful macro constants   */
/*:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

#ifndef LIMDEFS_H
#define LIMDEFS_H

/**
 * @addtogroup Libraries
 * @{
 */

#define MORGVERS "MORGAN Version 2.8.1" /* ** current MORGAN version ** */

#define MAX_PERR     200                /*Maximum no. of warnings or    */
                                        /*errors for parameter          */
                                        /*statements before quitting    */

#define MAX_WARNS    500                /*Maximum no. of warnings       */
                                        /*permitted from functions &    */
                                        /*programs before quitting      */

#define MAX_NCMP     200                /* initial value only for       */
                                        /* maximum no. of components    */
#define MAX_A_NCMP   200                /* same for Autozyg             */

#define FMAX_INT      20                /*Maximum no. of integers in    */
                                        /*pedigree file                 */
#define FMAX_FEF  FMAX_INT-2            /*Maximum no. of fixed effects  */
                                        /*in file plus mean             */
#define FMAX_REAL     20                /*Maximum no. of real no.'s     */
                                        /*(quant. traits) in pedfile    */
#define INT_CHR        6                /*Nominal number of characters  */
                                        /*for integers                  */
#define FLTPT_CHR      8                /*Nominal number of characters  */
                                        /*for real no.'s (quant. traits) */
                                        /*in pedigree file              */

#define MAX_RECSIZ  ((FMAX_INT)*(INT_CHR+2) + (FMAX_REAL)*(FLTPT_CHR+2))
                                        /*Maximum record size for       */
                                        /*pedigree file                 */

#define MIN_PED        3                /*Minimum pedigree size         */

#define MAX_U_PED    400                /*Maximum pedigree size for     */
                                        /*univar                        */
#define MAX_PED    20000                /*Maximum pedigree size for     */
                                        /*unibig, bivar, multivar       */
#define MAX_U_OBS    400                /*Maximum no. observeds for     */
                                        /*univar                        */
#define MAX_OBS    20000                /*Maximum no. observeds for     */
                                        /*univar, bivar, multivar       */
#define MAX_CPSIZ    400                /*Maximum component size for    */
                                        /*unibig, bivar, multivar       */

#define MAX_TR        10                /*Maximum no. of traits         */
                                        /*(all programs)                */
#define MAX_TRAIT      5                /*Upper bound on no. of         */
                                        /*quantitative traits           */
#define MAX_SIM_TR    10                /*Upper bound on no. of         */
                                        /*simulated traits              */
#define MAX_FEP       20                /*Upper bound on no. of differnt */
                                        /*fixed-effect parameter sets -- */
                                        /*e.g. one per chain in parallel */
                                        /*chain examples                */

#define MAX_CHROM     22                /*Maximum chromosome no.        */

#define MAX_MARK     500                /*Maximum no. of markers        */

#define MAX_LOC       500               /*Maximum no. of marker loci    */
                                        /*for Autozyg programs          */

#define MAX_ALLEL    128                /*Upper bound on no. of alleles */

                                        /* (p) (PedDatTyp)              */
                                        /* (t) (TraitTyp)               */
enum {
  ABSENT        = -9,                   /* (p)(t) No data or no trait loc */
  MARKER_GENO   =  1,                   /*    (t) Marker trait          */      
  GENOTYPIC     =  2,                   /* (p)(t) Genotypic code        */
  DISCRETE      =  3,                   /* (p)(t) Phenotypic            */
  ORDINAL       =  4,                   /* (p)                          */ 
  QUANTITATIVE  =  5,                   /* (p)(t) QTL                   */
  DISCR_LIAB    =  6,                   /* (p)    Discrete + liab class */
  INHERITANCE   =  7                    /* (p)(t) Pair of inherit indic */
   };

/** Maximum number of individuals to display */
#define MAX_N_DISP_INDI  50
/** Maximum number of loci to display */
#define MAX_N_DISP_LOCI  20
                                        /*for AUTOZYG and LODSCORE:     */
#define Q_SUMMN        0.9999           /*Lower and upper range for     */
#define Q_SUMMX        1.0001           /*Max sum of allele frequencies */
#define Q_LIMITS       ".9999, 1.0001"  /*(Unnormalized)                */

#define MARK_LOC       0                /*Marker locus                  */
#define TRAIT_LOC      1                /*Trait (tloc)  locus           */
#define UNSELMRK_LOC   2                /*Non-selected marker locus     */

#define UNKNOWNNAME    "0"
#define UNKNOWN        0                /*Unknown pedigree names are 0  */
                                        /*by convention                 */
#define MALE           1                /*"Usual" coding for sex        */
#define FEMALE         2
#define UNSEXED        0
#define AVERAGED       0

enum SEX
{ FEM, MAL, AVG };                      /*Subscripting for gender-      */
                                        /*specific arrays               */

enum MAP_TYP                            /*Choice of gender-specific,    */
{ SEX_SPEC, SEX_AVG, BOTH};             /*gender-averaged, or both maps */

#define UNOBSERVED     0                /*Code for "unobserved"         */
#define OBSERVED       1                /*for "observed" individual     */

                                        /********  for POLYEM:    *******/

#define FIXEFF         5                /*Number of classes of fixed    */
                                        /*effects, including mean       */
#define LEVELS        20                /*Maximum number of levels      */
                                        /*(categories) for each fixed   */
                                        /*effect                        */

#define ITMO          20                /*Breeding-value iterations     */

#define ITEM         200                /*Number of EM iterations       */

#define ITWT          50                /*How often to write out        */
                                        /*estimates                     */
#define DIFF           1.0e-8           /*Convergence criterion for     */
                                        /*breeding values iteration     */
                                        /*if = 0.0, does ITMO steps     */

                                        /******** for GENEDROP   ********/
                                        /*        (and others)          */
                                        /*Codes for gender, e.g:        */
#define UNDEF         -1                /*Operation is not specified    */
#define ASSIGN         1                /*Parameter is to be assigned   */
#define IGNORE_M       2                /*Parameter is to be ignored    */
#define CHECK_         3                /*Parameter is to be checked    */
#define ALL            4                /*Parameter is set to 'all'     */
#define USE            5                /*Parameter is to be used       */

#define SIGFIG         6                /*No. of significant figures    */
                                        /*for outputting trait values   */
#define MISSING      999                /*Trait value used for missing  */
                                        /*data                          */
#define MISSGENE       0                /*Genotype used for unobserved  */
                                        /*data                          */
#define MISSLABEL      0                /*Label used for unobserved     */
                                        /*data                          */
#define MA1ST          1                /*Mother comes 1st in record    */
#define PA1ST          2                /*Father comes 1st in record    */

                                        /***** for AUTOZYG:         *****/
enum { 
  INDEP_LOC,                            /* Use locus-by-locus sampling,       */
  SEQIMP                                /* or sequential imputation for setup */
   };

#define RSEQIMP       .10               /*Default no. of sequential     */
                                        /*imputation samples            */
                                        /* RSEQIMP*NIt                  */
                                        /*Default no. burnin iterations: */
#define RBRNIT_MAP      .10             /*   RBRNIT_MAP * NIt           */
#define RBRNIT_PVAL     .10             /*   RBRNIT_PVAL * NIt          */
#define RBRNIT_IBDTESTS .10             /*   RBRNIT_IBDTESTS * NIt      */
#define RBRNIT_AUTO     .10             /*   RBRNIT_AUTO * NIt          */
#define RBRNIT_LODS    1.0              /*   RBRNIT_LODS * NIt          */
#define RBRNIT_BAYES    .20             /*   RBRNIT_BAYES * NPreIt      */
#define RBRNIT_MRKS     .10             /*   RBRNIT_MRKS * NIt          */
#define RBRNIT_SCHNELL  .10             /*   RBRNIT_SCHNELL * NIt       */
#define RBRNIT_TWOQTL   .10             /*   RBRNIT_TWOQTL * NIt        */
                                        /*Default pseuto-prior iteratns */
#define RPREIT          .50             /*   RPREIT * NIt               */

#define MCEMIT       100                /*Default no. MCEM iterations   */ 

#define NC             7                /*Maximum # of components       */

#define NOBS         300                /*Maximum # observed individuals */

#define NLOC          24                /*Maximum # markers             */

#define MAXGAM        10                /*Max 10 gametes                */

/* PTgg is no. of IBD patterns for n gametes  (may need +1)             */
/* CDgg is max. code for n gametes (labels start with 0)                */

#ifdef DEFINE_DATA
/* Number of gametes:0  1  2  3   4   5    6     7     8      9      10 */

GLOBAL int PTgg[11] = { 0, 0, 2, 5, 15, 52, 203, 877, 4140, 21147, 115975 };
GLOBAL int CDgg[11] = { 0, 0, 1, 4, 15, 64, 325, 1956, 13699, 109600, 986409 };

/* Default beginning and ending trait positions                         */
GLOBAL int NXtrnPos = 5;
GLOBAL double XtrnPos[5] = { .03, .07, .15, .25, .4 };

/* Default interval proportions for trait                               */
GLOBAL int NIntProp = 3;
GLOBAL double IntProp[3] = { .2, .5, .8 };

#else
GLOBAL int PTgg[11];
GLOBAL int CDgg[11];

GLOBAL int NXtrnPos;
GLOBAL double XtrnPos[5];
GLOBAL int NIntProp;
GLOBAL double IntProp[3];

#endif /*DEFINE_DATA */
                                        /*********  for PARS:    ********/

#define MAXLINLEN    2000               /*The max # of characters in an  */
                                        /*parameter statement line       */
#define PMAXLINLEN  "2000"              /*Same as above, but for printng */

#define WHITESPACE  " ,\t\v\n\f\r"      /*"whitespace" in the parameter  */
                                        /*statements                     */
#define NUL          '\0'               /*null character for end-of-line */

#define COMMENT       '#'               /*The rest of the line is a      */
                                        /*comment                        */
#define ENDPAR     "****"               /*Marks end of parameter         */
                                        /*statements in parameter file   */

enum DISPOPT 
{ NO_DISP, DEFLT_DISP, SEGR_DISP, LABL_DISP }; /*Graphics display options */

#define INVALID_INDEX -1

#define MAXNAMLEN      15               /* max # of chars in a name     */

#define N              0
#define Y              1
#define FALSE          0
#define TRUE           1
#define OK             1

#ifndef M_PI
#define M_PI           3.14159265358979323846
#endif
#define PPI            (2*M_PI)     /* 2-pi for use in Normal densities */

/* Define uint32 to be a 32 bit unsigned integer type                   */
/* (used by Rans/ and Pars/ functions                                   */
/* If there isn't one, the code would have to be rewritten              */
/* to emulate 32 bit arithmetic.                                        */

typedef unsigned int uint32;            /*this works on RISC workstations */

#define CONGRUENTIAL_SEED 12345u        /*default seed values           */
#define TAUSWORTHE_SEED   1073u         /*for random no. generator      */

/** @} */

#endif /*LIMDEFS_H */
