/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*
 * @file   pedchk.h
 * @author MAJ 
 * @date   2000
 *
 * @brief Data structures used by pedchk functions also function prototypes
 * for Pedchk functions.
 *
 * \$Id: pedchk.h,v 1.15 2005/08/14 00:19:36 thompson Exp $
 */

#ifndef PEDCHK_H
#define PEDCHK_H

/**
 * @defgroup pedchk libpedchk
 * @ingroup Libraries
 * @{
 */

#include "incs.h"

                                        /* For writing:                       */
GLOBAL int widPNam;                     /* max length of names                */
GLOBAL int minPInt[FMAX_INT-3],         /* smallest and                       */
           maxPInt[FMAX_INT-3],         /* largest integer                    */
           widPInt[FMAX_INT-3];         /* no. characters                     */

GLOBAL int nCompnt;                     /* no. components in pedigree         */
GLOBAL int nInd;                        /* no. individuals found in pedigree  */
GLOBAL int nObs;                        /* no. observed individuals in ped    */
GLOBAL int nFound;                      /* no. founders in pedigree           */
GLOBAL int nFGL;                        /* no. founder genome labels in ped   */

GLOBAL int nGend[3];                    /* no. females, males, unsexed        */

                        /*compntList:  pedigree components                    */
struct compntList {
    int nInd;
    int nObs;
    int nFound;
    int nFGL;
    int nGend[3];
    char frstNam [MAXNAMLEN + 1];
                } ;
typedef struct compntList COMPS;
typedef COMPS * COMPS_PTR;
GLOBAL COMPS_PTR pCompTab;

                        /*indKeys:  key table for individuals                 */
struct indKeys {
    char nam  [MAXNAMLEN+1];            /* input name                         */
    int ix;                             /* input record no. (0, 1, ...)       */
    int comp;                           /* component no. (1, 2, ...)          */
    int chron;                          /* chronological no. (1, 2, ...)      */
    int founder;                        /* founder?                           */
    int ixMa;                           /* record no. for mother              */
    int ixPa;                           /* record no. for father              */
               };
typedef struct indKeys KEYS;
typedef KEYS * KEYS_PTRS;
typedef KEYS_PTRS * KEYS_PP;
GLOBAL KEYS_PTRS pFirstKeyTab;
GLOBAL KEYS_PP ppKeyTab;

                        /*pedDat:  pedigree data input for indivduals         */
struct pedDat {
    char nam  [MAXNAMLEN+1];
    char namMa[MAXNAMLEN+1];
    char namPa[MAXNAMLEN+1];
    int sex;
    int obsrvd;
              };
typedef struct pedDat PEDD;
typedef PEDD * PED_PTR;
GLOBAL PED_PTR pPedDat;

                        /*intDat:  optional integer data in input pedigree    */
                        /*         file for individuals                       */
typedef int * PEDI_PTR;
GLOBAL PEDI_PTR pIntDat;

                        /*RSDat:  optional real data from input pedigree      */
                        /*        file for individuals                        */
typedef char * PEDRS_PTR;
typedef PEDRS_PTR * PEDRS_PP;
GLOBAL PEDRS_PP ppRSDat;
GLOBAL int maxRSDat;                    /*no. characters allocated            */
GLOBAL int curRSDat;                    /*no. characters used                 */


void   allocPed(void),            reallocRSDat(void),   /* ped.c              */
       writChkPed(FILE *),        freeChkPed(void);
int    readPed(FILE *),           checkPed(FILE *),
       assignChkPed(void);
FILE * openChkPed(void);

void   allocKey(void),            storKey(void),        /* key.c              */
       freeKey(void);

int    checkChron(void),          newChron(void);       /* chron.c            */
void   assignChron(void),         reassignIx(void),
       reassignChron(void);

void   allocComp(void),           assignComp(void),     /* comp.c             */
       storComp(void),            prntComp(void);

int    checkNam(void);                                  /* nam.c              */

void   findParents(void);                               /* parent.c           */
int    checkParents(void);
int    unknown (char *);

void   assignGender(void);                              /* gender.c           */
int    checkGender(void);

void   sortChron(void),           sortNam(void),        /* sort.c             */
       sortIx(void),              sortComp(void);
int    comparNam(const void *, const void *),
       comparIx(const void *, const void *),
       comparChron(const void *, const void *),
       comparComp(const void *, const void *);

/* int    comparNam(const void *, const void *),       *** moved to stuff.h   */

void   chkPedFil(void);
int    chk_pedfil(void),          mak_opedfil(void);


/** @} */

#endif /*PEDCHK_H*/
