/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*******************************************************************************
  program name: lm_markers.c (lm_lods lange-gamma-version -- renamed lm_markers)
             mow modified for quantitative traits -- file renamed lm_quant

  Program to do lod scores (replaces homozyg.c from old Autozyg)
  The state (X) consists of segregation indicators -- so can generalize to
        cases of more than one observed/pedigree, and multiple loci.

  The data, Y, is the multilocus genotypes of observed individuals
     in the pedigree.

  Obtained by modification of jnt_auto.c.  December 1998.  (EAT)

Here is the plan!
    OK this is the PLAN  (12/15/98)

OK, now we have a general autozyg program, we need to go back to
lod estimation, for those as want.  We use the same program (modified).

Since the data are locus-by-locus, the likelihood ratio contribution is
just ratio of P_theta(S)/P_sim_th(S).

This is very easy, once we get the chromosome set-ups fixed.

1) Trait locus data set-up -- assume trait-locus genotypes for observed
individuals -- done  12/20/98
2) Likelihood-ratio contributions routine,  -- done 12/18/98
3) Simulation and evaluation chromosomes for each location. -- done 12/23/98

Now to integrate the L-sampler, using the new Lsampler and Msampler
      libraries and new routines;        EAT March 2000

Converted to MORGAN_V2.3                 MAJ June 2000

Modifications to specify simulation points, within each interval and
outside of map.    EAT, then converted to MORGAN 2.3 by MAJ, Nov 2000.

Modification of full_seg_prob, and recombinant counts     EAT  Nov 2000
    scoring of counts and tot_log_probs by component and locus/interval.

Modifications to lik_rat, do_lsmpl, etc.                  EAT  Dec 2000

Modifications to use specify intervals for l-ratio updating and use
    more efficient seg_prob updating routines for l-sample steps
                                                          EAT Jan 2001

Added changes for sequential imputation set-up            AWG March 2001

But now we can use locus_mjrpeel with sample_flag zero
     to do a Lange-style lod score using only sampled S_M    EAT May 2001
  We can use the extval and intval positions as computation positions
    -- except we can also compute at the loci, maybe??

This is the version to test new trait stuctures.
Program renamed lm_markers.                                 EAT March 2002

Version modified for quantitative traits (lm_quant.c)       EAT Jan 2004

Output scores file added for CDLL output                    EAT Apr 2004

*******************************************************************************/

/**
 * @defgroup lm_markers lm_markers
 * @ingroup lodscore
 * @{
 */

#define PNAM  "lm_markers"
#define MARKERS

#include "lodscore.h"
#include "setupAutozyg.c"

int main (int argc, char *argv[])
{
   int   j, k, jj;
   int   t, c, h, nc;
   int   ntp;
   int   nsw, nh;
   int   bsz, bc;                       /* btch-size = NIt/NBtch,
                                           bc = current batch                 */

   /* these base_lik and rb_ arrays now contain tot (element 0) and by compnt */
   double **rb_prob, ***rb_prbm;        /* cumulative rb_likelihood
                                           and batch totals by position       */
   double **rb_sd;                      /* Monte Carlo SDevs                  */
   double *base_lik;                    /* base-point log P(Y_T)              */
   double *cmpt_lik;                    /* log-lik by Component               */

   double *ctot;                        /* CDLL by component                  */
   int ***recounts;
   double **prob;
   /* likelihoods etc.        */
   double **tot_log_prob_seg;
   double **tot_log_prob_pen;
   double **temp_log_prob_seg;
   int   tot_count;

   /* double cumupos[2]; */

   /* Map position (cM) of current test position */
   double markpos[2];                   /* Map position (cM) of current marker
                                         ** NOTE unused */
   int   lastmarker = 0;                /* Last test position
                                         ** NOTE unused */
   double **rec_left;                   /* fem =0, male=1; by test position   */
   double **rec_rght;                   /* fem =0, male=1; by test position   */
   int  *loc_left;                      /* locus left of test position        */
   int  *loc_rght;                      /* locus rght of test position        */

   FILE *oScorF, *oSeedF;

   /* .................................................. */

   nsw = setupAutozyg (argc, argv, PNAM, 0);
   prob = setupInherit (1);

   /* .................................................. */

   ctot = malloc ((nCompnt + 1) * sizeof (double));
   alloc_chk (WHERE, ctot);

   alloc_loglik (&recounts, &tot_log_prob_seg, &temp_log_prob_seg,
                 &tot_log_prob_pen);

   /* now set evaluation points for lange_lik */
   /* ntp is number of trait positions plus number of markers (NLocPos + 1) */

   alloc_lf_rt (NLocPos + 1, &rec_left, &rec_rght, &loc_left, &loc_rght);

   ntp = set_eval_points (NMarks, loc_left, loc_rght,
                          rec_left[0], rec_rght[0], rec_left[1], rec_rght[1]);

   if (ntp != NLocPos + 1)
      give_up_where (WHERE, ERA_mrk, NLocPos, ntp);

   /* moved down initial peeling after alloc_rb etc : EAT July 2004 */

   if (nCompnt == 1)
      PeelCmpnt = 0;

   alloc_rb (ntp, NBtch, nCompnt, PeelCmpnt, &base_lik, &cmpt_lik,
             &rb_prob, &rb_prbm, &rb_sd);

   nc = PeelCmpnt * nCompnt + 1;

   for (c = 0; c < nc; ++c)
   {
      base_lik[c] = 0.0;
      cmpt_lik[c] = 0.0;
      for (nh = 0; nh < ntp; ++nh)
      {
         rb_prob[c][nh] = 0.0;
         rb_sd[c][nh] = 0.0;
         for (jj = 0; jj < NBtch; ++jj)
            rb_prbm[c][jj][nh] = 0.0;
      }
   }

   /* Added EAT -- P(Y_T) */
   /* set_transmission (loci, 0, NMarks);   */

   set_simple_trans (loci, -1, -1, 0.5, 0.5, 0.5, 0.5);

   if (TraitTyp != QUANTITATIVE)
      base_lik[0] = locus_mjrpeel (1.0, 1.0, 0, PeelCmpnt, cmpt_lik, loci);
   else
      base_lik[0] = locus_mjrpeel (loci[0].ve, 1.0, 0, PeelCmpnt, cmpt_lik,
                                   loci);

   /*   checks for Zero probability in peeling operation                      */
   if (base_lik[0] == -DBL_MAX)
      give_up_where (WHERE, ERS_zpp);

   printf (" Base log-likelihood P(Y_T) = %e\n \n", base_lik[0]);
   fflush (stdout);

   if (PeelCmpnt)
   {
      printf (" Base log-likelihoods by component = \n   ");
      for (c = 1; c <= nCompnt; ++c)
      {
         base_lik[c] = cmpt_lik[c];
         printf (" %e ", base_lik[c]);
      }
      printf (" \n \n");
      fflush (stdout);
   }

   if ((NIt / ScorIt) * ScorIt != NIt)
      give_warn (ERR_err,
                 "Scoring interval does not divide realizations count");
   tot_count = NIt / ScorIt;

   bsz = tot_count / NBtch;
   /* checks for no. batches dividing scored iterations */
   if (bsz * NBtch != tot_count)
      give_warn (ERL_bt1, NBtch);
   /* checks for batch size > 0 */
   if (!bsz)
      give_up (ERL_bt2, NBtch);

   printf (" %d realizations will be scored, at scoring interval %d\n",
           tot_count, ScorIt);
   printf (" Scored realizations will be batched in %d sets of size %d", NBtch,
           bsz);
   printf ("\n \n");
   fflush (stdout);

   for (c = 0; c < nCompnt; ++c)
   {
      for (h = 0; h <= NMarks; ++h)
      {
         tot_log_prob_seg[c][h] = 0.0;
         tot_log_prob_pen[c][h] = 0.0;
      }
      for (j = 0; j <= NMarks; ++j)
      {
         for (jj = 0; jj < 6; ++jj)
            recounts[c][j][jj] = 0;
      }
   }
   tot_count = 0;                       /* reset tot_count to count when
                                           scored */
   bc = 0;                              /* current batch indicator */

   fflush (stdout);

   alloc_logprob ();

   /* open output scores file, if requested */

   if (FilIt)
      if (!(oScorF = open_fil (OSCORE)))
         give_up (ERR_qit);

   prntStatusBMC ();                    /* print status for beginning MCMC */

   /* now do burn-in iterations */

   printf (" Now starting %d iterations of burn-in\n", NBrnIt);
   fflush (stdout);

   for (t = 1; t <= NBrnIt; t++)
   {
      if (!(t % ChkIt))
      {
         printf (" Starting burn-in iteration %d\n", t);
         fflush (stdout);
      }
      sampling (t - 1, 1, NMarks, nsw, nFGL, prob);
   }

   /* now do MCMC sampling, and score results */

   NMarks = NMarks;

   printf (" Now starting %d iterations of MCMC\n", NIt);
   fflush (stdout);
   for (t = 0; t < NIt; ++t)
   {
      if (DebugMain)
         dbgXXGG ();

      /* MCMC sampling step or scan */
      sampling (t, 1, NMarks, nsw, nFGL, prob);

      if (!((t + 1) % ScorIt))
      {
         /* score tot log-probs and recombs */
         for (c = 0; c < nCompnt; ++c)
         {
            ctot[c] = 0.0;
            full_seg_prob (NMarks, c, recounts, temp_log_prob_seg);
            for (j = 1; j < NMarks; ++j)
            {
               tot_log_prob_pen[c][j] += log (prob[c][j]);
               tot_log_prob_seg[c][j] += temp_log_prob_seg[c][j];
               ctot[c] += (log (prob[c][j]) + temp_log_prob_seg[c][j]);
            }
            tot_log_prob_pen[c][NMarks] += log (prob[c][NMarks]);
            ctot[c] += log (prob[c][NMarks]);
         }

         /* score rb_likelihoods  */
#if 0
         /* zero out meiosis indicators for line-by-line check */
         for (i = 0; i < PedSize; ++i)
            if (nghd[i].pa != &nullnode)
               for (j = 0; j <= NMarks; ++j)
               {
                  xx0[j][i] = 0;
                  xx1[j][i] = 0;
               }
#endif

         /* start new batch if current btch is complete */

         if (bc < NBtch)
            do_rblik (ntp, bc, PeelCmpnt, loc_left, loc_rght, rec_left[1],
                      rec_left[0], rec_rght[1], rec_rght[0], rb_prob, rb_prbm,
                      base_lik, cmpt_lik);
         ++tot_count;

         if (!((tot_count) % bsz))
            ++bc;
      }
      if (FilIt && !((t + 1) % FilIt))  /* tests for output to scores fil */
         for (c = 0; c < nCompnt; ++c)
         {
            fprintf (oScorF, "%d %d %f", t + 1, c + 1, ctot[c]);
            fprintf (oScorF, "\n");
         }
   }
   printf ("\n");

   prntStatusFMC ();                    /* print status for finished with mcmc */

   /* compute likelihood estimates  */

   for (c = 0; c < nc; ++c)
      for (k = 1; k < ntp; ++k)
         /* if (loc_left[k] != loc_rght[k])      */
      {
         rb_prob[c][k] = rb_prob[c][k] / (NBtch * bsz);
         for (bc = 0; bc < NBtch; ++bc)
            rb_prbm[c][bc][k] = rb_prbm[c][bc][k] / bsz;
      }

   /* Batch-means standard devation estimates  */
   /* careful with component indexing in arrays !! If PeelCmpnt, nc=nCompnt+1 */

   for (c = 0; c < nc; ++c)
      for (k = 1; k < ntp; ++k)
      {
         /*   if (loc_left[k] != loc_rght[k])                   */
         {
            for (bc = 0; bc < NBtch; ++bc)
               rb_sd[c][k] += (rb_prbm[c][bc][k] - rb_prob[c][k])
                  * (rb_prbm[c][bc][k] - rb_prob[c][k]);
            rb_sd[c][k] = sqrt (rb_sd[c][k] / ((NBtch - 1) * NBtch));
         }
      }

   /* output results for this simulation.estimation set */
   /* Moved down so liks can also be by compnt: EAT July 2004  */

   for (c = 0; c < nCompnt; ++c)
   {
      pr_cmpnthead (c);

      pr_log_prob_seg (1, NMarks, -cmplist[c + 1].nmeioses, tot_log_prob_seg[c],
                       tot_count);

      pr_log_prob_pen (1, NMarks, tot_log_prob_pen[c], tot_count);

      pr_reco_counts (1, NMarks, recounts[c]);

      if (PeelCmpnt)
      {
         prntLodSHead ();

         for (k = 1; k < ntp; ++k)
         {
            prntLodSPos (k - 1);
            printf (" %11.4f%10.4f\n", log10 (rb_prob[c + 1][k]),
                    rb_sd[c + 1][k] / (rb_prob[c + 1][k] * log (10.0)));
         }
         printf ("\n");
         fflush (stdout);

      }
   }
/* print cumulative lhood ests */

   if (PeelCmpnt)
      prntLodSJHead ();
   else
      prntLodSHead ();

   for (k = 1; k < ntp; ++k)
   {
      if ((!prntLodSPos (k - 1)) && (TraitTyp == GENOTYPIC))
         /* the current test position is a marker position. 
            This test is reinstated: EAT Oct 2005.   It is needed for
            GENOTYPIC traits in some cases.  See note re do_rblik() */
         printf (" %11s%10s\n", "NA", "NA");
      else
         printf (" %11.4f%10.4f\n", log10 (rb_prob[0][k]),
                 rb_sd[0][k] / (rb_prob[0][k] * log (10.0)));
   }
   printf ("\n");
   fflush (stdout);

   /* now we are done with MCMC */

   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }

   /* Free up allocated memory */
   free_xxgg ();
   free_allefrq ();
   free_chromo (0);
   free_allocate_loci ();
   free_perm ();
   locus_peel_free ();

   /* now we are done  */

   exit (EXIT_SUCCESS);
}
