/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*************************************************************************

  This file contains the functions to do the mixed model segregation
  part of the joint linkage and segregation analyses. (GLS June '99)

  Reorganized to get rid of mixd_init(), global variable z now becomes
  'additive' defined in main() of schnell.c and is passed to smpl_a through
  the last argument.  "head.h" seems to be no longer needed and is not
  included.                                                (NL, July 01)

  $Id: mixds2.c,v 1.5 2006/03/24 19:02:18 myrna Exp $

 **************************************************************************/

#include "mixd.h"

void smpl_a (struct Locus *loci, int n_loci, double ve, double va,
             double mu, double mynp, double *zadd)
{
   /* Global variables used */

   int   gt;
   double y;

   /* indices */
   int   i = 0;
   int   j = 0;

   /* substract off global mean
    * note both ym and xtrait start index at 1 and go to Nq
    */

   for (i = 0; i < PedSize; ++i)
   {
      if (nghd[i].q_dat)
      {
/* NOTE nghd.q_dat stores the polygene-adjusted  trait value  */
         nghd[i].q_dat[0] += zadd[i];
         y = nghd[i].q_dat[0];          /* Original trait value */
         for (j = 0; j < n_loci; ++j)
         {
            if (loci[j].locus_type == TRAIT_LOCUS)
            {
               gt = loci[j].gt[0][i] + loci[j].gt[1][i] - 2;
               y -= loci[j].g_mu[gt];     /***  maybe also -mu  ***/
            }
         }
         cond_z (i, y, ve, va, zadd);
      }
/* BUG FIX: EAT Aug 2004: polygenes must be sampled for unobserved also!! */
      else
         cond_z (i, 0.0, ve, va, zadd);
   }

   for (i = 0; i < PedSize; ++i)
   {
      if (nghd[i].q_dat)
         nghd[i].q_dat[0] -= zadd[i];
   }
}

void cond_z (int i, double y_mu, double ve, double va, double *zadd)
{
   /* Global variables used */
   extern Ind *nghd;                    /* nghds.h */

   double mean, sd, s;
   int   no_kids;
   Ind  *dad, *mum, *spouse, *kid;
   int   dad_indx, mum_indx, spouse_indx, kid_indx;
   Marlist *mar_ptr;
   Indlist *kid_ptr;

   dad = nghd[i].pa;

   if (!(isNullNode (dad)))
   {
      mum = nghd[i].ma;
      dad_indx = dad->index;
      mum_indx = mum->index;
      sd = 2.0 / va;
      mean = (zadd[dad_indx] + zadd[mum_indx]) / va;
   }
   else
   {
      sd = 1.0 / va;
      mean = 0.0;
   }
   if (nghd[i].q_dat != NULL)
   {                                    /* not missing */
      mean += y_mu / ve;
      sd += 1.0 / ve;
   }

   mar_ptr = nghd[i].marriages;
   while (mar_ptr != NULL)
   {                                    /* has at least one spouse */
      no_kids = 0;
      s = 0.0;

      spouse = mar_ptr->current->spouse;
      spouse_indx = spouse->index;

      kid_ptr = mar_ptr->current->kids;

      while (kid_ptr != NULL)
      {                                 /* has at least one kid */

         ++no_kids;
         kid = kid_ptr->current;
         kid_indx = kid->index;
         s += zadd[kid_indx];

         kid_ptr = kid_ptr->next;       /* next kid, if any */
      }

      sd += 0.5 * no_kids / va;
      mean += (s - 0.5 * no_kids * zadd[spouse_indx]) / va;

      mar_ptr = mar_ptr->next;          /* next spouse, if any */
   }

   mean /= sd;
   sd = sqrt (1.0 / sd);
   zadd[i] = normal (mean, sd);
}

/* 
 * $Log: mixds2.c,v $
 * Revision 1.5  2006/03/24 19:02:18  myrna
 * Global variable np is renamed Nq
 *
 * Revision 1.4  2005/08/14 00:20:32  thompson
 * Adding copyright statement at file head.
 *
 * Revision 1.3  2005/03/16 07:32:34  myrna
 * README_newSetup contains pseudo-code for the new setupLodSc() function.
 * This function call replaces set-up code in the lodscore programs.
 * In addition, some redundant variables are replaced, and some function
 * calls are changed.
 *
 * Revision 1.2  2004/08/20 02:59:25  thompson
 * Schnell is updated to use new nghd trait-data structures, and get rid
 * of old arrays (qdex, ym, x, xtrait, etc.).
 * ALSO IMPORTANT NOTE:
 * schnell was WRONG: failing to sample polygenic values of unobserved individuals
 *           -- a substantive logic bug is fixed in smpl_a in mixd2.c
 *           -- old gold standard checked out (pre bug fix). New golds committed.
 * NOTE ALSO:
 * schnell is still BAD:  polygenic updates are still by single-site updating
 *       -- needs to be fixed (one day) by using polypeel to peel and resample.
 *
 * Revision 1.1  2004/04/22 23:17:50  myrna
 * lm_schnell is added to Lodscores/ directory
 *
 * Revision 1.4  2002/02/13 06:43:37  thompson
 * Updated alloc_routines in schnell.c, and founder checking in mixd2.c
 *
 * Revision 1.3  2001/09/13 20:00:37  nali
 *
 * 1.  Call 'free_allocate' corresponding to 'allocate_loci'.
 * 2.  Remove schnell
 * 3.  Remove file scope variable 'iobs2' (read_gen_dat2.c) and local nobs
 * (read_gen_dat()).  Use global varibales NObs, and IObs[NOBS]
 * 4.  Call allocate_loci() then do necessary post-processing.  Move
 * allocate()instead.variable)global variable 'gene_idx' which doesn't seem
 * 5.  Introduce a set of functions for allocating arrays, and use them.into
 * read_gen_dat2().)to be used.
 *
 */
