/*  vim: sw=3
 *  $Id: print_lods.c,v 1.24 2006/03/24 19:04:40 myrna Exp $ */

static int nCol1;                       /* file global for formatting
                                         * column headings and data columns */

#include "markers.h"

void print_lods (double ***lkratios)
{
   /* Global variables used */
   extern int n_compnts;
   extern int NTPos;

   double **lods = NULL;                /* Lod scores */
   double **lods_cmb = NULL;            /* Combined Lod Scores */
   double lodsval[4];
   double eigen2 = 0.0;
   int   c, i, m;

   lods = alloc_dbl2d (3, NTPos);
   lods_cmb = alloc_dbl2d (3, NTPos);

   printf ("\n\n ESTIMATED LOD SCORES\n\n");

   for (c = 0; c < n_compnts; ++c)
   {
      printf ("  Component %d\n", c + 1);
      for (m = 0; m < 3; ++m)
      {
         lodsval[m] = comput_lods (lkratios[c], lods[m], NTPos, m, &eigen2);
         if (m == 0)
            lodsval[3] = eigen2;
      }
      printf ("\n    The largest eigenvalue:        %9.5f\n", lodsval[0]);
      printf ("\n    The second largest eigenvalue: %9.5f\n", lodsval[3]);
      printf ("\n    Cumulative from left  :        %9.5f\n", lodsval[1]);
      printf ("\n    Cumulative from right :        %9.5f\n", lodsval[2]);
      printf ("\n    LOD scores:\n");
      printf ("  %10s %10s %10s %10s %10s %10s\n", "position",
              "female", "male", "eigen", "left", "right");
      for (i = 0; i < NTPos; ++i)
      {
         printf ("  %10d", i);
         printf (" %10.5f %10.5f",
                 map_position (i, FEM), map_position (i, MAL));
         for (m = 0; m < 3; ++m)
         {
            lods_cmb[m][i] += lods[m][i];
            printf (" %10.5f", lods[m][i]);
         }
         putchar ('\n');
      }
      putchar ('\n');
      fflush (stdout);
   }
   free_2d ((void *) &lods, 3);

   if (n_compnts > 1)
   {
      printf ("  Combined All Components:\n");
      printf ("  %10s %10s %10s %10s %10s %10s\n", "position",
              "female", "male", "eigen", "left", "right");
      for (i = 0; i < NTPos; ++i)
      {
         printf ("  %10d", i);
         printf (" %10.5f %10.5f",
                 map_position (i, FEM), map_position (i, MAL));
         for (m = 0; m < 3; ++m)
            printf (" %10.5f", lods_cmb[m][i]);
         putchar ('\n');
      }
   }
   putchar ('\n');
   fflush (stdout);

   free_2d ((void *) &lods_cmb, 3);
}
void prntLodSHead (void)
{
   nCol1 = LocNamLen > 11 ? LocNamLen : 11;

   if (IS_LM_MARKERS)
      printf ("\n LodScore estimates by Rao-Blackwellized computation:\n\n");

   if (IS_LM_LODS)
      printf ("\n LodScore estimates:\n\n");

   if (IS_LM_BAYES)
      printf ("\n LodScore estimates:\n\n");

   if (IS_LM_TWOQTL)
      printf ("\n LodScore estimates:\n\n");

   prntLodSColHd ();

   return;
}
void prntLodSJHead (void)
{
   printf ("\n LodScores estimated jointly over all components:\n\n");

   prntLodSColHd ();

   return;
}

void prntLodSColHd (void)
{
   printf (" %*s     position (Haldane cM)", nCol1, "Trait pos #");

   if (IS_LM_BAYES)
      printf ("%11s %8s %19s", "pseudo", "freq", "LodScore    ");

   printf ("\n");

   printf (" %*s        male     female   ", nCol1, "or marker");

   if (IS_LM_MARKERS)
      printf (" %11s%10s", "LodScore", "StdErr");

   if (IS_LM_LODS)
      printf (" %10s %10s %10s ", "eigen", "left", "right");

   if (IS_LM_BAYES)
      printf ("%11s %8s %9s %9s", "prior", "visited", "crude", "R-B");

   if (IS_LM_TWOQTL)
      printf (" %11s%10s", "LodScore", "StdErr");

   printf ("\n\n");
   fflush (stdout);

   return;
}

/**
 * prints trait position no. or marker name, and chrom position for ea gender.
 * returns true for trait position, false for markers position
 */
int prntLodSPos (int j)
{
   static int ntp, nmk;
   static LOCPOS_PTR p_p;
   int   t;

   nCol1 = LocNamLen > 11 ? LocNamLen : 11;

   if (!j)
   {
      p_p = LocPos;
      ntp = 0;
      nmk = 0;
   }
   if (p_p->locus_typ == TRAIT_LOC || p_p->locus_typ == UNSELMRK_LOC)
   {
      ++ntp;
      printf (" %*d", nCol1, ntp);
      t = 1;
      if (p_p->locus_typ == TRAIT_LOC)
         printf (" ");
      else if (p_p->locus_typ == UNSELMRK_LOC)
         printf ("*");

   }
   else if (p_p->locus_typ == MARK_LOC)
   {
      printf (" %*s ", nCol1, MarkNam[nmk]);
      ++nmk;
      t = 0;
   }
   printf ("%11.3f%11.3f   ", p_p->pos[MAL], p_p->pos[FEM]);

   ++p_p;

   return t;
}
void prntLodSPos0 (void)
{
   printf (" %*d", nCol1, 0);
   printf (" %11s%11s   ", "unlinked", "unlinked");

   return;
}
