/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*************************************************************************
 kin.c

 This program combines sim_ibd_link.c and autolink.c, described below

 This program is implemented in MORGAN_V2.4.1, with parameter statements
 added for kin.

 For each pedigree component:
   kinship coefficients are computed for pairs of individuals
   single-locus inbreeding coeffs are computed for specified individuals
   2-locus inbreeding coeffs are computed for specified individuals.

 Usage of this program is described in the README_userdoc for this
 directory.                                               May 2002  MAJ

**************************************************************************
  sim_ibd_link.c

  Program to simulate IBD  to estimate probs
  The state (X) consists of segregation indicators --
  X determines the IBD pattern among the specified descendant gametes.
       Up to 10 gametes can be scored for IBD,

  Two methods of scoring IBD -- by pattern marginally over loci,
      and by a specified (group of) pattern(s) over windows.

  Original program modified from jntauto,                Fall 1999  EAT
  New scoring routines added                             Fall 1999
  Combined into single program,                           May 2000  EAT

  Cleaning up and reformatting the code.
  Unused variables deleted and pointers initialized.
  Can now be built with MORGAN V2.3 libraries.            Aug 2000  N.L.

**************************************************************************
 autolink.c

 Program copied from autonew.c in Twoloc/Utahprogs/Autozyg directory,
 implementing the recursive routine for gene identity at two loci.

 This then determines the probabilities of each of the 2^k patterns of
 gene identity at k loci, in a single inbred individual.  -- assumed
 MARKOV IBD -- omit this part!! - use only 2-locus ibd computation.

 g4link rountine corrected,                               Mar 2000  EAT

 Rewritten so that it can now read in standard pedigree data used by
 other MORGAN 2.3 programs and also can use the same data structure and
 kinship() function in the Nghd library.
                                                          Aug 2000  N.L.

 Modified to call by index rather than name-id            Oct 2001  N.L.

 Modified for several probands, and improved input and output
                                                          Oct 2001  EAT
*************************************************************************/

#define PNAM "kin"
#include "pedcomp.h"
#include "pedcomp_fncs.c"

/**
 * @defgroup kin kin
 * @ingroup PedComp
 * @{
 */

int main (int argc, char *argv[])
{
   int   c, i, j, wid;
   double x, g4;

   Ind  *pr;

   setupKin (argc, argv, PNAM);

   set_kinix (nCompnt);

   printf ("\n");

   for (c = 0; c < NKinCmp; ++c)
   {
      if (NKinNams[KINSHP][c] || NKinNams[INBRD][c] || NKinNams[INBRD2][c])
      {
         printf (" Component %d:\n\n", c + 1);

         if (NKinNams[KINSHP][c])
         {
            printf ("    Kinship coefficients:\n\n");
            for (i = 0; i < NKinNams[KINSHP][c]; i += 2)
            {
               x = kinship (nghd + KinIx[KINSHP][c][i],
                            nghd + KinIx[KINSHP][c][i + 1]);
               printf ("    %*s  %*s %s\n", widPNam, KinNam[KINSHP][c][i],
                       widPNam, KinNam[KINSHP][c][i + 1], dub2nlz (x, 8, 5));
            }
            printf ("\n\n");
         }
         if (NKinNams[INBRD][c])
         {
            printf ("    Inbreeding coefficients:\n\n");
            for (i = 0; i < NKinNams[INBRD][c]; ++i)
            {
               x = inbreeding (nghd + KinIx[INBRD][c][i]);
               printf ("    %*s %s\n", widPNam, KinNam[INBRD][c][i],
                       dub2nlz (x, 8, 5));
            }
            printf ("\n\n");
         }
         if (NKinNams[INBRD2][c])
         {
            wid = widPNam < 7 ? 7 : widPNam;
            printf ("    2-locus inbreeding coefficients:\n");
            printf ("    (g4link is probability of IBD at both of 2 linked"
                    " loci)\n\n");

            printf ("    %*s  recomb   g4link\n", wid, "proband");
            printf ("    %*s    freq     prob\n", wid, " ");
            for (i = 0; i < NKinNams[INBRD2][c]; ++i)
            {
               pr = nghd + KinIx[INBRD2][c][i];
               if (pr->pa == &nullnode)
                  printf ("\n    %*s     is founder\n", wid,
                          KinNam[INBRD2][c][i]);
               else if (inbreeding (pr) == 0.)
                  printf ("\n    %*s     is not inbred\n", wid,
                          KinNam[INBRD2][c][i]);
               else
                  for (j = 0; j < NInbreedRF; ++j)
                  {
                     if (!j)
                        printf ("\n    %*s", wid, KinNam[INBRD2][c][i]);
                     else
                        printf ("    %*s", wid, " ");
                     g4 = g4link (pr, 0, pr, 1, pr, 0, pr, 1, InbreedRF[j]);
                     printf ("  %s", dub2nlz (InbreedRF[j], 6, 3));
                     printf (" %s\n", dub2nlz (g4, 8, 5));
                  }
            }
            printf ("\n\n");
         }
      }
      fflush (stdout);
   }
   if (InbreedRF)
      free (InbreedRF);

   free_auto ();

   exit (EXIT_SUCCESS);
}

void set_kinix (int nc)
{
   int   k, c, i, posp;

   for (k = 0; k < 3; ++k)
      for (c = 0; c < NKinCmp; ++c)
         for (i = 0; i < NKinNams[k][c]; ++i)
         {
            if ((posp = get_posit (KinNam[k][c][i])) == INVALID_INDEX)
               give_up (ERA_inp, KinNam[k][c][i]);
            else if ((nghd[posp].component) != (c + 1))
               give_up (ERA_pnc, KinNam[k][c][i], c + 1, nghd[posp].component);

            KinIx[k][c][i] = posp;
         }

   if (NKinCmp < nc)
      give_warn (ERK_klc, NKinCmp);
   else if (NKinCmp > nc)
      give_warn (ERK_kmc, NKinCmp);

   return;
}

/** @} */
