/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

void  setupKin (int, char **, char *);
void  initLodSc (char **);
void  procKinPars (int, char **, char *);
void  chkPedFil (void);
double **setupQPed (void);
void  setPed (double **);

void setupKin (int argc, char **argv, char *ppn)
{

   /* initialize exit procedure, print headings. 
    * fncs:  atexit(), print_head() 
    * dirs:  Stuff 
    */
   initLodSc (argv);

   /* process parameter statements, set up globals resulting from these.
    * fncs:  proc_pars(), prntStatusPP(), proc_auto(), proc_fils(), 
    *        proc_seeds(), prntStatusCP() 
    * vars:  globals in globs.h
    * dirs:  Pars, Stuff
    */
   procKinPars (argc, argv, ppn);

   /* check, and massage if necessary, pedigree file.
    * fncs:  chk_pedfil(), prntStatusCPF(), mak_opedfil()
    * vars:  indKeys[PedSize], compntList[nCompnt]
    * dirs:  PedComp, Stuff
    */
   chkPedFil ();

   /* read in pedigree data and set up pedigree structure.
    * fncs:  alloc_ped(), open_ped(), input_ped(), count_founders() 
    * vars:  int data[PedSize][NColData], Ind nghd[PedSize+1]
    *        double quant[NTraits][PedSize], Ind nghd[].g_dat[2*(NMarks+1)]
    * dirs:  Nghds, Quant, Stuff
    */
   setPed (NULL);

   return;
}

/* ........................................................................ */

/* the functions below are to be moved to appropriate directories  */

/**
 * Initialize exit procedure;  print header.
 */
void initLodSc (char **argv)
{
   atexit (print_tail);

   print_head (*argv);

   return;
}

/**
 * Process the parameter file and set corresponding globals
 * (these start with capital letters, and are found in globs.h)
 */
void procKinPars (int argc, char **argv, char *ppn)
{

   /* parse input parameter statements  */
   if (!proc_pars (argc, argv, ppn))
      exit (EXIT_FAILURE);

   prntStatusPP ();

   /* check statements for completeness and consistency.
    * Set up global parameters and data structures     */
   if (!proc_kin ())
      exit (EXIT_FAILURE);

   /* set up pedigree file parameters */
   if (!proc_fils ())
      exit (EXIT_FAILURE);

   /* set up seed parameters */
   if (!proc_seeds ())
      exit (EXIT_FAILURE);

   prntStatusCP ();

   return;
}

/**
 * read and check pedigree file;  write new one if needed.
 */
void chkPedFil (void)
{
   /* check pedigree file for errors */
   if (!chk_pedfil ())
      give_up (ERR_qit);

   prntStatusCPF ();

   /* write new pedigree file, if changes to file */
   if (!mak_opedfil ())
      give_up (ERR_qit);

   return;
}

/**
 * Read and store pedigree data.
 */
void setPed (double **qdata)
{
   int   n;
   FILE *pedF;

   /* allocate structures for pedgree data:  int data[PedSize][NColData]
    *                                        Ind nghd[PedSize + 1]  */
   alloc_ped ();

   /* open for reading */
   if (!(pedF = open_ped ()))
      give_up (ERR_qit);

   /* read in pedigree and trait data:       data[][]  */
   input_ped (pedF, qdata);

   fclose (pedF);

   /* set up neighborhood and pedigree components data structures */
   nghds ();

   /* do an extra check on no. founders */
   if ((n = count_founders ()) != nFound)
      give_up_where (WHERE, ERN_cfd, nFound, n);

   /* set up lists of members in each component and also 
      information about each component */
   if ((n_compnts = components ()) != nCompnt)
      give_up_where (WHERE, ERN_ccp, nCompnt, n_compnts);

   return;
}
