/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*  functions are:
 *        assignComp()
 *        allocComp()
 *        storComp()
 *        prntComp()
 */
#include "pedchk.h"

void assignComp (void)
{
   int   nPedDone = 0;
   int   nThisLoop;
   KEYS_PP ppK;

   nCompnt = 0;

   do                                   /* finds members of next component */
   {
      ++nCompnt;
      for (ppK = ppKeyTab; ppK < ppKeyTab + nInd &&
           (!(*ppK)->founder || (*ppK)->comp); ++ppK)
         ;                              /* finds next founder not assigned * 
                                         * to a component                  */
      (*ppK)->comp = nCompnt;           /* assigns this founder to comp    */
      ++nPedDone;

      do
      {
         nThisLoop = 0;

         for (ppK = ppKeyTab; ppK < ppKeyTab + nInd; ++ppK)
         {
            /*                           * checks for no assigned comp,    *
             *                           * but parent in this comp         */
            if (!(*ppK)->comp && !(*ppK)->founder)
            {
               if (((*(ppKeyTab + (*ppK)->ixMa - 1))->comp) == nCompnt
                   || ((*(ppKeyTab + (*ppK)->ixPa - 1))->comp) == nCompnt)
               {
                  if (!((*(ppKeyTab + (*ppK)->ixMa - 1))->comp))
                  {
                     (*(ppKeyTab + (*ppK)->ixMa - 1))->comp = nCompnt;
                     ++nPedDone;
                     ++nThisLoop;
                  }
                  if (!((*(ppKeyTab + (*ppK)->ixPa - 1))->comp))
                  {
                     (*(ppKeyTab + (*ppK)->ixPa - 1))->comp = nCompnt;
                     ++nPedDone;
                     ++nThisLoop;
                  }
                  (*ppK)->comp = nCompnt;
                  ++nPedDone;
                  ++nThisLoop;
               }
            }
            /*                           * checks for in this comp,        *
             *                           * but unassigned parent(s)        */
            else if ((*ppK)->comp == nCompnt && !(*ppK)->founder)
            {
               if (!((*(ppKeyTab + (*ppK)->ixMa - 1))->comp))
               {
                  (*(ppKeyTab + (*ppK)->ixMa - 1))->comp = nCompnt;
                  ++nPedDone;
                  ++nThisLoop;
               }
               if (!((*(ppKeyTab + (*ppK)->ixPa - 1))->comp))
               {
                  (*(ppKeyTab + (*ppK)->ixPa - 1))->comp = nCompnt;
                  ++nPedDone;
                  ++nThisLoop;
               }
            }
         }
      }
      while (nThisLoop);
   }
   while (nPedDone < nInd);

   return;
}
void allocComp (void)
{
   pCompTab = (COMPS_PTR) malloc (nCompnt * sizeof (COMPS));
   alloc_chk (WHERE, pCompTab);

   return;
}
void storComp (void)
{
   KEYS_PP ppK;
   COMPS_PTR pC;

   int   i = -1;

   pC = pCompTab - 1;

   for (ppK = ppKeyTab; ppK < ppKeyTab + nInd; ++ppK)
   {
      if ((*ppK)->comp > i + 1)         /* inits table for next compnt */
      {
         ++i;
         ++pC;
         pC->nInd = 0;
         pC->nObs = 0;
         pC->nFound = 0;
         pC->nFGL = 0;
         pC->nGend[UNSEXED] = 0;
         pC->nGend[MALE] = 0;
         pC->nGend[FEMALE] = 0;

         strncpy (pC->frstNam, (*ppK)->nam, MAXNAMLEN + 1);
      }
      ++pC->nInd;                       /* counts members of component, */
      if ((*ppK)->founder)              /* founders and founder genome labels */
      {
         ++pC->nFound;
         pC->nFGL += 2;
      }
      if (ObsIn && (pPedDat + (*ppK)->ix - 1)->obsrvd != UNOBSERVED)
         ++pC->nObs;                    /* .. and observed, if known   */
      ++pC->nGend[(pPedDat + (*ppK)->ix - 1)->sex]; /* and gendrs */
   }

   pC = pCompTab;

   for (i = 0; i < nCompnt; ++i, ++pC)  /* finds totals for all compnt */
   {
      nObs += pC->nObs;
      nFound += pC->nFound;
      nGend[UNSEXED] += pC->nGend[UNSEXED];
      nGend[MALE] += pC->nGend[MALE];
      nGend[FEMALE] += pC->nGend[FEMALE];
   }
   nFGL = 2 * nFound;

   return;
}
void prntComp (void)
{
   int   i, wid;
   COMPS_PTR pC;

   pC = pCompTab;

   wid = widPNam < 7 ? 7 : widPNam;

   if (nCompnt == 1)
   {
      fprintf (stdout, " Pedigree information:\n\n");

      fprintf (stdout, "    size  founders  females    males");
      if (nGend[UNSEXED])
         fprintf (stdout, "  unsexed");
      if (nObs)
         fprintf (stdout, "  observd");
      fprintf (stdout, "  %*s", wid, "first");
      fprintf (stdout, "\n\n");
      fprintf (stdout, " %7d   %7d  %7d  %7d", pC->nInd,
               pC->nFound, pC->nGend[FEMALE], pC->nGend[MALE]);
      if (nGend[UNSEXED])
         fprintf (stdout, "  %7d", pC->nGend[UNSEXED]);
      if (nObs)
         fprintf (stdout, "  %7d", pC->nObs);
      fprintf (stdout, "  %*s\n", wid, pC->frstNam);
   }
   else
   {
      fprintf (stdout, " Pedigree component information:\n\n");

      fprintf (stdout, "    compnt     size  founders  females    males");
      if (nGend[UNSEXED])
         fprintf (stdout, "  unsexed");
      if (nObs)
         fprintf (stdout, "  observd");
      fprintf (stdout, "  %*s", wid, "first");
      fprintf (stdout, "\n\n");

      for (i = 0; i < nCompnt; ++i, ++pC)
      {
         fprintf (stdout, "    %6d  %7d   %7d  %7d  %7d", i + 1, pC->nInd,
                  pC->nFound, pC->nGend[FEMALE], pC->nGend[MALE]);
         if (nGend[UNSEXED])
            fprintf (stdout, "  %7d", pC->nGend[UNSEXED]);
         if (nObs)
            fprintf (stdout, "  %7d", pC->nObs);
         fprintf (stdout, "  %*s\n", wid, pC->frstNam);
      }
      fprintf (stdout, "\n     total");
      fprintf (stdout, "  %7d  %7d  %7d  %7d",
               nInd, nFound, nGend[FEMALE], nGend[MALE]);
      if (nGend[UNSEXED])
         fprintf (stdout, "  %7d", nGend[UNSEXED]);
      if (nObs)
         fprintf (stdout, "  %7d", nObs);
      fprintf (stdout, "\n");
   }
   fprintf (stdout, "\n");

   return;
}
