/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/****************************************************************************
 *                                                                          *
 *     Joint peeling routines for trait loci extracted from Loki            *
 *                                                                          *
 *     Simon Heath - The Rockefeller University                             *
 *                                                                          *
 *                    April 1998                                            *
 *                                                                          *
 * lsample_peel_init.c:                                                     *
 *                                                                          *
 * (1) Determine peeling sequence for trait locus                           *
 *                                                                          *
 * (2) Allocate stordage for peeling operations                             *
 *                                                                          *
 ****************************************************************************/

#include "lsample.h"

static int *allele_trans, *ngenos, *gene_obs, *n_rfuncs;
static struct R_Func **r_func;
static int n_prev_inv, comp, r_func_size, *involved, *prev_inv, *true_involved,
   *rf_flag;
static unsigned long *temp_set, *id_set[2], mask, **all_set, *req_set[3];
struct Family **id_fam, *famlist, **flist;
int  *idflag, max_peel_off, n_singles, *singles, *nfam;
int  *cplist, *cp_size;
char *MMsg = "Out of Memory Error - aborting\n";

/* Flag families 'surrounding' (i.e., neighbours to) fam */
#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "FlagFam"
static void FlagFam (const struct Family *fam)
{
   int   k, ids, idd;
   Marlist *mar;
   Indlist *kids;

   ids = fam->sire->index;
   idd = fam->dam->index;
   if (id_fam[ids])
      id_fam[ids]->flag |= 1;           /* If sire appears as offspring, 
                                           flag that family */
   mar = fam->sire->marriages;          /* And all the families for which 
                                           he is a parent */
   while (mar)
   {
      k = mar->current->kids->current->index;
      id_fam[k]->flag |= 1;
      mar = mar->next;
   }
   /* Do the same for the dam */
   if (id_fam[idd])
      id_fam[idd]->flag |= 1;
   mar = fam->dam->marriages;
   while (mar)
   {
      k = mar->current->kids->current->index;
      id_fam[k]->flag |= 1;
      mar = mar->next;
   }
   /* And for each kid */
   kids = fam->off;
   while (kids)
   {
      mar = kids->current->marriages;
      while (mar)
      {
         k = mar->current->kids->current->index;
         id_fam[k]->flag |= 1;
         mar = mar->next;
      }
      kids = kids->next;
   }
}

static void print_orig_alleles (FILE * fptr, Ind * id, struct Locus *locus)
{
   int   i, j;

   i = id->index;
   j = locus->data_index;
   if ((id->g_dat) && (id->g_dat[2 * j]))
      fprintf (fptr, "%d,%d", id->g_dat[2 * j], id->g_dat[2 * j + 1]);
   else
      fputs ("*,*", fptr);
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "print_alls"
static void print_alls (FILE * fptr, Ind * id, const int n_all)
{
   int   j, k, k1, l;
   unsigned long a, *b, m;

   if (!(b = calloc (n_all, sizeof (unsigned long))))
      ABT_FUNC (MMsg);
   j = id->index;
   m = (1 << n_all) - 1;
   for (k1 = k = 0; k < n_all; k++)
   {
      a = all_set[k][j];
      if (a == m)
         k1++;
      for (l = 0; l < n_all; l++)
         if (a & (1 << l))
            b[l] |= 1 << k;
   }
   if (k1 == n_all)
   {
      fputs ("(*,*)", fptr);
   }
   else
   {
      for (k = 0; k < n_all; k++)
      {
         l = 0;
         a = all_set[k][j];
         if (a == m)
            fprintf (fptr, "(*,%d) ", 1 + allele_trans[k]);
         else
            while (a)
            {
               if (b[l] != m && (a & 1))
                  fprintf (fptr, "(%d,%d) ", 1 + allele_trans[l],
                           1 + allele_trans[k]);
               l++;
               a >>= 1;
            }
      }
      for (l = 0; l < n_all; l++)
         if (b[l] == m)
            fprintf (fptr, "(%d,*) ", 1 + allele_trans[l]);
   }
   fputc ('\n', fptr);
   free (b);
}

static void PrintFamily (struct Family *fam, const int n_all,
                         struct Locus *locus)
{
   Indlist *kids;

   fprintf (stderr, "Father: %s [", fam->sire->self);
   print_orig_alleles (stderr, fam->sire, locus);
   fputs ("]\n        ", stderr);
   print_alls (stderr, fam->sire, n_all);
   fprintf (stderr, "Mother: %s [", fam->dam->self);
   print_orig_alleles (stderr, fam->dam, locus);
   fputs ("]\n        ", stderr);
   print_alls (stderr, fam->dam, n_all);
   kids = fam->off;
   while (kids)
   {
      fprintf (stderr, "     -> %s [", kids->current->self);
      print_orig_alleles (stderr, kids->current, locus);
      fputs ("]\n        ", stderr);
      print_alls (stderr, kids->current, n_all);
      kids = kids->next;
   }
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "DoFamily"
static int DoFamily (struct Family *fam, const int n_all, struct Locus *locus)
{
   int   i, j, k, l, ids, idd, nc = 0, fg, kid, change = 0;
   unsigned long a, b, cm;
   Indlist *kids;

   ids = fam->sire->index;
   idd = fam->dam->index;
   for (i = 0; i < n_all; i++)
      temp_set[i] = id_set[0][i] = id_set[1][i] = 0;
   for (i = 0; i < n_all; i++)
   {
      a = all_set[i][ids];
      if (a)
         for (j = 0; j < n_all; j++)
            if (a & (1 << j))
            {
               if (j > i && (all_set[j][ids] & (1 << i)))
                  continue;
               for (k = 0; k < n_all; k++)
               {
                  b = all_set[k][idd];
                  if (b)
                     for (l = 0; l < n_all; l++)
                        if (b & (1 << l))
                        {
                           if (l > k && (all_set[l][idd] & (1 << k)))
                              continue;
                           cm = ((1 << i) | (1 << j));
                           fg = 0;
                           kids = fam->off;
                           while (kids)
                           {
                              kid = kids->current->index;
                              if (!((all_set[k][kid] & cm) ||
                                    (all_set[l][kid] & cm)))
                              {
                                 fg = 1;
                                 break;
                              }
                              kids = kids->next;
                           }
                           if (fg)
                              continue;
                           temp_set[k] |= cm;
                           temp_set[l] |= cm;
                           id_set[0][i] |= 1 << j;
                           id_set[0][j] |= 1 << i;
                           id_set[1][k] |= 1 << l;
                           id_set[1][l] |= 1 << k;
                           nc++;
                        }
               }
            }
   }
   for (k = l = j = 0; j < n_all; j++)
   {
      if (all_set[j][ids] & id_set[0][j])
         k++;
      if (all_set[j][idd] & id_set[1][j])
         l++;
      if (k && l)
         break;
   }
   if (!nc || j == n_all)
   {
      printf ("%d %d\n", nc, j);
      fputs ("\nDoFamily(): Inconsistent family data\n", stderr);
      PrintFamily (fam, n_all, locus);
      return 1;
   }
   for (j = 0; j < n_all; j++)
   {
      a = all_set[j][ids];
      all_set[j][ids] &= id_set[0][j];
      if (a != all_set[j][ids])
         change = 1;
      a = all_set[j][idd];
      all_set[j][idd] &= id_set[1][j];
      if (a != all_set[j][idd])
         change = 1;
   }
   kids = fam->off;
   while (kids)
   {
      kid = kids->current->index;
      l = n_all;
      for (j = 0; j < l; j++)
      {
         cm = temp_set[j];
         if (all_set[j][kid] & cm)
            break;
      }
      if (j == l)
      {
         fputs ("\nDoFamily(): (Internal error?) Inconsistent family data\n",
                stderr);
         PrintFamily (fam, n_all, locus);
         return 1;
      }
      for (j = 0; j < l; j++)
      {
         a = all_set[j][kid];
         cm = temp_set[j];
         all_set[j][kid] &= cm;
         if (a != all_set[j][kid])
            change = 1;
      }
      kids = kids->next;
   }
   if (change)
      FlagFam (fam);
   fam->flag &= ~1;
   return 0;
}

void calc_order (const int nfam, struct Family **flist, int *order)
{
   int   i;
   Ind  *sire, *dam;
   Indlist *id_list;

   for (i = 0; i < nfam; i++)
   {
      sire = flist[i]->sire;
      dam = flist[i]->dam;
      order[sire->index]++;
      order[dam->index]++;
      id_list = flist[i]->off;
      while (id_list)
      {
         order[id_list->current->index]++;
         id_list = id_list->next;
      }
   }
   for (i = 0; i < pedsize; i++)
      if (ngenos[i] == 1)
         order[i] = 1;
}

static int check_haps (const int i)
{
   int   fgs = 0;

   if (idflag[i] & PEELED)
      return fgs;
   fgs |= HAP_MAT | HAP_PAT;
   if (gene_obs[i] && ngenos[i] > 1)
      fgs |= HAP_JNT;
   if ((nghd[i].obs) || (idflag[i] & WAS_PIVOT))
      fgs |= HAP_JNT;
   return fgs;
}

static void set_flags (const int nfam, struct Family **flist, const int comp,
                       struct R_Func *rf)
{
   int   i, j, k, fm, fp, kid;
   Indlist *id_list;
   Ind  *sire, *dam;

   id_list = cmplist[comp].people;
   if (id_list)
      while (id_list->next)
      {
         id_list = id_list->next;
         idflag[id_list->current->index] &= STABLE_FLAGS;
      }
   for (i = 0; i < nfam; i++)
   {
      sire = flist[i]->sire;
      dam = flist[i]->dam;
      fp = check_haps (sire->index);
      fm = check_haps (dam->index);
      id_list = flist[i]->off;
      while (id_list)
      {
         kid = id_list->current->index;
         if (!(idflag[kid] & PEELED))
         {
            idflag[kid] |= check_haps (kid);
            idflag[kid] |= HAP_M | HAP_P;
         }
         id_list = id_list->next;
      }
      fm |= (HAP_MAT | HAP_PAT | HAP_JNT);
      fp |= (HAP_MAT | HAP_PAT | HAP_JNT);
      idflag[sire->index] |= fp;
      idflag[dam->index] |= fm;
   }
   for (i = 0; i < n_rfuncs[comp - 1]; i++)
      if (!rf[i].flag)
      {
         for (j = 0; j < rf[i].n_ind; j++)
         {
            k = abs (rf[i].id_list[j]) - 1;
            idflag[k] |= (HAP_MAT | HAP_PAT | HAP_JNT);
         }
      }
}

static void add_to_involved (const int i, int *n, int flag)
{
   int   j;

   for (j = 0; j < (*n); j++)
      if (i == involved[j])
         break;
   if (j == (*n))
   {
      rf_flag[j] = 0;
      involved[(*n)++] = i;
   }
   if (flag)
      rf_flag[j] = 1;
}

static int assemble_matrix (int size, int *g_perm, int *iw, int *pe, int *len,
                            int *trans, int k2)
{
   int   i, x, i1, j, k, l, kid, n_inv, iwlen;
   Marlist *marlist;
   Indlist *id_list;
   iwlen = 0;

   for (x = 0; x < size; x++)
   {
      i1 = g_perm[x];
      i = abs (i1) - 1;
      n_inv = 0;
      add_to_involved (i1, &n_inv, 0);
      if (idflag[i] & HAP_JNT)
         add_to_involved (-i1, &n_inv, 0);
      if (i1 < 0)
      {                                 /* Put in parents */
         if (idflag[i] & HAP_P)
         {
            add_to_involved (nghd[i].pa->index + 1, &n_inv, 0);
            add_to_involved (-1 - nghd[i].pa->index, &n_inv, 0);
         }
      }
      else
      {
         if (idflag[i] & HAP_M)
         {
            add_to_involved (nghd[i].ma->index + 1, &n_inv, 0);
            add_to_involved (-1 - nghd[i].ma->index, &n_inv, 0);
         }
      }
      marlist = nghd[i].marriages;
      while (marlist)
      {
         if (marlist->current->kids->current->pa == nghd + i)
            l = 1;
         else
            l = 0;
         id_list = marlist->current->kids;
         while (id_list)
         {
            kid = id_list->current->index;
            if (l)
            {
               if (idflag[kid] & HAP_P)
                  add_to_involved (-(kid + 1), &n_inv, 0);
            }
            else if (idflag[kid] & HAP_M)
               add_to_involved (kid + 1, &n_inv, 0);
            id_list = id_list->next;
         }
         marlist = marlist->next;
      }
      for (j = 0; j < n_rfuncs[comp - 1]; j++)
         if (!r_func[comp - 1][j].flag)
         {
            for (k = 0; k < r_func[comp - 1][j].n_ind; k++)
               if (r_func[comp - 1][j].id_list[k] == i1)
                  break;
            if (k < r_func[comp - 1][j].n_ind)
            {
               for (k = 0; k < r_func[comp - 1][j].n_ind; k++)
               {
                  l = r_func[comp - 1][j].id_list[k];
                  add_to_involved (l, &n_inv, 0);
               }
            }
         }
      if (iw)
      {
         pe[x] = iwlen + 1;
         for (k = 1; k < n_inv; k++)
            iw[iwlen++] = trans[k2 + involved[k]];
         len[x] = n_inv - 1;
      }
      else
         iwlen += n_inv - 1;
   }
   return iwlen;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "min_degree"
static int *min_degree (int *g_perm, int size)
{
   int   x, i1, k1, k2, *trans;
   int  *pe, *iw, *degree, *nv, *next, *last, *head, *elen, *w, *len;
   int   iwlen, pfree, ncmpa, iovflo;

   k2 = INT_MAX;
   k1 = INT_MIN;
   for (x = 0; x < size; x++)
   {
      i1 = g_perm[x];
      if (i1 < k2)
         k2 = i1;
      if (i1 > k1)
         k1 = i1;
   }
   k2 = -k2;
   k1 += k2 + 1;
   if (!(trans = malloc (sizeof (int) * k1)))
      ABT_FUNC (MMsg);
   for (x = 0; x < k1; x++)
      trans[x] = -1;
   for (x = 0; x < size; x++)
   {
      i1 = g_perm[x] + k2;
      trans[i1] = x + 1;
   }
   iwlen = assemble_matrix (size, g_perm, 0, 0, 0, 0, 0);
   if (iwlen + size > (int) (iwlen * 1.2))
      iwlen += size;
   else
      iwlen = (int) (iwlen * 1.2);
   if (!(last = malloc (sizeof (int) * size)))
      ABT_FUNC (MMsg);
   if (!(pe = malloc (sizeof (int) * (8 * size + iwlen))))
      ABT_FUNC (MMsg);
   degree = pe + size;
   nv = degree + size;
   next = nv + size;
   head = next + size;
   elen = head + size;
   w = elen + size;
   len = w + size;
   iw = len + size;
   pfree = 1 + assemble_matrix (size, g_perm, iw, pe, len, trans, k2);
   iovflo = INT_MAX;
   amdbar_ (&size, pe, iw, len, &iwlen, &pfree, nv, next, last, head, elen,
            degree, &ncmpa, w, &iovflo);
   free (pe);
   free (trans);
   return last;
}

static int count_ops (const int i, const int mflag, const int flag)
{
   int   i1, j, k, l, m, kid, n_inv = 1;
   Marlist *marlist;
   Indlist *id_list;

   i1 = mflag ? -(i + 1) : i + 1;
   involved[0] = i1;
   rf_flag[0] = 0;
   if (idflag[i] & HAP_JNT)
   {
      rf_flag[n_inv] = 0;
      involved[n_inv++] = -i1;
   }
   if (mflag)
   {                                    /* Put in parents */
      if (idflag[i] & HAP_P)
      {
         add_to_involved (nghd[i].pa->index + 1, &n_inv, 0);
         add_to_involved (-1 - nghd[i].pa->index, &n_inv, 0);
         if (flag)
            idflag[i] |= HAD_P;
      }
   }
   else
   {
      if (idflag[i] & HAP_M)
      {
         add_to_involved (nghd[i].ma->index + 1, &n_inv, 0);
         add_to_involved (-1 - nghd[i].ma->index, &n_inv, 0);
         if (flag)
            idflag[i] |= HAD_M;
      }
   }
   marlist = nghd[i].marriages;
   while (marlist)
   {
      if (marlist->current->kids->current->pa == nghd + i)
         l = 1;
      else
         l = 0;
      id_list = marlist->current->kids;
      while (id_list)
      {
         kid = id_list->current->index;
         if (l)
         {
            if (idflag[kid] & HAP_P)
            {
               add_to_involved (-(kid + 1), &n_inv, 0);
               if (flag)
                  idflag[kid] ^= (HAP_P | HAD_P);
            }
         }
         else
         {
            if (idflag[kid] & HAP_M)
            {
               add_to_involved (kid + 1, &n_inv, 0);
               if (flag)
                  idflag[kid] ^= (HAP_M | HAD_M);
            }
         }
         id_list = id_list->next;
      }
      marlist = marlist->next;
   }
   for (j = 0; j < n_rfuncs[comp - 1]; j++)
      if (!r_func[comp - 1][j].flag)
      {
         for (k = 0; k < r_func[comp - 1][j].n_ind; k++)
            if (r_func[comp - 1][j].id_list[k] == i1)
               break;
         if (k < r_func[comp - 1][j].n_ind)
         {
            for (k = 0; k < r_func[comp - 1][j].n_ind; k++)
            {
               l = r_func[comp - 1][j].id_list[k];
               add_to_involved (l, &n_inv, 1);
            }
            if (flag)
               r_func[comp - 1][j].flag = 1;
         }
      }
   for (j = 0; j < n_rfuncs[comp - 1]; j++)
      if (!r_func[comp - 1][j].flag)
      {
         for (k = 0; k < r_func[comp - 1][j].n_ind; k++)
         {
            m = r_func[comp - 1][j].id_list[k];
            if (m == i1)
               break;
            for (l = 0; l < n_inv; l++)
               if (m == involved[l])
                  break;
            if (l == n_inv)
               break;
         }
         if (flag && k == r_func[comp - 1][j].n_ind)
            r_func[comp - 1][j].flag = 1;
      }
   if (flag)
   {
      if (mflag)
         idflag[i] &= ~(HAP_P | HAP_PAT | HAP_JNT);
      else
         idflag[i] &= ~(HAP_M | HAP_MAT | HAP_JNT);
   }
   else if (n_prev_inv)
   {
      for (j = 0; j < n_inv; j++)
      {
         l = involved[j];
         for (k = 0; k < n_prev_inv; k++)
            if (l == prev_inv[k])
               break;
         if (k == n_prev_inv)
            break;
      }
      if (j == n_inv)
         return -1;
   }
   return n_inv - 1;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "jntpeel_sequence"
static void jntpeel_sequence (int n_genes, int *order, int *g_perm,
                              struct Peelseq_Head *pp)
{
   int   x, i, j, k, k1 = -1, k2, k3, k4, l, pivot, n_ops, n_rf;
   int   n_inv = 0, n_peel;
   struct Complex_Element *element;

   for (x = 0; x <= n_genes; x++)
   {
      if (x == n_genes)
         pivot = n_ops = 0;
      else
      {
         pivot = order[x];
         i = g_perm[pivot - 1];
         if (i > 0)
            n_ops = count_ops (i - 1, 0, 0);
         else
            n_ops = count_ops (-1 - i, 1, 0);
      }
      if (n_inv && n_ops >= 0)
      {
         n_peel = n_inv;
         for (k = 0; k < n_prev_inv; k++)
         {
            j = prev_inv[k];
            true_involved[n_inv++] = j;
         }
         if (!(element = malloc (sizeof (struct Complex_Element))))
            ABT_FUNC (MMsg);
         pp->type = PEEL_COMPLEX;
         pp->ptr.complex = element;
         pp = &element->next;
         pp->type = 0;
         element->n_peel = n_peel;
         element->n_involved = n_inv;
         for (n_rf = k = 0; k < n_rfuncs[comp - 1]; k++)
            if (r_func[comp - 1][k].flag == 1)
               n_rf++;
         element->n_rfuncs = n_rf;
         if (!(element->involved = malloc (sizeof (int) * (n_inv * 2 + n_rf))))
            ABT_FUNC (MMsg);
         element->flags = element->involved + n_inv;
         element->index = element->flags + n_inv;
         if (n_peel == n_inv)
         {
            free (r_func[comp - 1][k1].id_list);
            n_rfuncs[comp - 1]--;
            k1 = -1;
         }
         element->out_index = k1;
         for (k = 0; k < n_inv; k++)
            element->flags[k] = 0;
         for (k = 0; k < n_inv; k++)
         {
            l = true_involved[k];
            if (l > 0)
            {
               element->flags[k] |= idflag[l - 1] & HAD_M;
               idflag[l - 1] &= ~HAD_M;
            }
            else
            {
               element->flags[k] |= idflag[-1 - l] & HAD_P;
               idflag[-l - 1] &= ~HAD_P;
            }
         }
         for (j = 0; j < n_inv; j++)
            element->involved[j] = true_involved[j];
         for (n_rf = k = 0; k < n_rfuncs[comp - 1]; k++)
            if (r_func[comp - 1][k].flag == 1)
            {
               for (k1 = 0; k1 < r_func[comp - 1][k].n_ind; k1++)
               {
                  for (j = 0; j < n_inv; j++)
                     if (r_func[comp - 1][k].id_list[k1] == true_involved[j])
                        break;
                  if (j == n_inv)
                     ABT_FUNC ("Internal error - allele not found.\n");
                  r_func[comp - 1][k].id_list[k1] = j;
                  element->flags[j] |= IN_RF;
               }
               element->index[n_rf++] = k;
               r_func[comp - 1][k].flag = 2;
            }
         for (k = 0; k < n_inv - 1; k++)
         {
            l = true_involved[k];
            for (j = k + 1; j < n_inv; j++)
               if (true_involved[j] == -l)
               {
                  for (k2 = 0; k2 < n_rfuncs[comp - 1]; k2++)
                     if (r_func[comp - 1][k2].flag == 1)
                     {                  /* Check if both alleles are in 
                                         *same* R-Func */
                        for (k3 = k4 = 0; k3 < r_func[comp - 1][k2].n_ind; k3++)
                        {
                           if (l == r_func[comp - 1][k2].id_list[k3])
                              k4++;
                           if (-l == r_func[comp - 1][k2].id_list[k3])
                              k4++;
                           if (k4 == 2)
                              break;
                        }
                        if (k4 == 2)
                           break;
                     }
                  if (k2 == n_rfuncs[comp - 1])
                  {
                     element->flags[j] |= HAP_JNT;
                     element->flags[k] |= HAP_JNT;
                  }
                  break;
               }
         }
         k1 = element->out_index;
         if (k1 >= 0)
            for (k = 0; k < r_func[comp - 1][k1].n_ind; k++)
               r_func[comp - 1][k1].id_list[k] = true_involved[n_peel + k];
         n_inv = 0;
         n_prev_inv = 0;
      }
      if (!pivot)
         break;
      j = pivot - 1;
      pivot = g_perm[j];
      if (pivot > 0)
         n_ops = count_ops (pivot - 1, 0, 1);
      else
         n_ops = count_ops (-1 - pivot, 1, 1);
      if (!n_inv)
      {
         if (n_rfuncs[comp - 1] >= r_func_size)
         {
            r_func_size <<= 1;
            if (!(r_func[comp - 1] = realloc (r_func[comp - 1],
                                              sizeof (struct R_Func) *
                                              r_func_size)))
               ABT_FUNC (MMsg);
         }
         k1 = n_rfuncs[comp - 1]++;
         if (!(r_func[comp - 1][k1].id_list =
               malloc (sizeof (int) * (1 + n_ops))))
            ABT_FUNC (MMsg);
         r_func[comp - 1][k1].n_ind = n_ops;
         r_func[comp - 1][k1].flag = 0;
         r_func[comp - 1][k1].n_terms = 0;
         for (k = 1; k <= n_ops; k++)
            r_func[comp - 1][k1].id_list[k - 1] = involved[k];
      }
      else if (n_inv)
      {
         for (k = 0; k < r_func[comp - 1][k1].n_ind; k++)
            if (r_func[comp - 1][k1].id_list[k] == pivot)
            {
               r_func[comp - 1][k1].id_list[k] =
                  r_func[comp - 1][k1].id_list[--r_func[comp - 1][k1].n_ind];
               break;
            }
         r_func[comp - 1][k1].flag = 0;
      }
      for (k = 1; k <= n_ops; k++)
         prev_inv[k - 1] = involved[k];
      n_prev_inv = n_ops;
      true_involved[n_inv++] = pivot;
   }
   for (k1 = i = 0; i < n_rfuncs[comp - 1]; i++)
      if (!r_func[comp - 1][i].flag)
      {
         fputs ("Unused R-Function on: ", stderr);
         for (j = 0; j < r_func[comp - 1][i].n_ind; j++)
         {
            k = r_func[comp - 1][i].id_list[j];
            print_orig_allele_id (stderr, k);
            fputc (' ', stderr);
         }
         fputc ('\n', stderr);
         k1 = 1;
      }
   if (k1)
      ABT_FUNC ("Aborting\n");
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "find_sequence"
static void find_sequence (const int nfam, struct Family **flist,
                           struct Peelseq_Head *pp, struct R_Func *rf)
{
   int   i, n_genes = 0;
   int  *g_perm, *order, *order1;
   Indlist *id_list;

   set_flags (nfam, flist, comp, rf);
   id_list = cmplist[comp].people;
   if (id_list)
      while (id_list->next)
      {
         id_list = id_list->next;
         i = id_list->current->index;
         if (idflag[i] & HAP_MAT)
            n_genes++;
         if (idflag[i] & HAP_PAT)
            n_genes++;
      }
   if (!n_genes)
      return;
   if (!(involved = malloc (sizeof (struct R_Func) * n_genes * 6)))
      ABT_FUNC (MMsg);
   n_prev_inv = 0;
   prev_inv = involved + n_genes;
   true_involved = prev_inv + n_genes;
   rf_flag = true_involved + n_genes;
   g_perm = rf_flag + n_genes;
   order1 = g_perm + n_genes;
   n_genes = 0;
   id_list = cmplist[comp].people;
   if (id_list)
      while (id_list->next)
      {
         id_list = id_list->next;
         i = id_list->current->index;
         if (idflag[i] & HAP_MAT)
            g_perm[n_genes++] = i + 1;
         if (idflag[i] & HAP_PAT)
            g_perm[n_genes++] = -(i + 1);
      }
   order = min_degree (g_perm, n_genes);
/*      for(i=0;i<n_genes;i++) order[i]=i+1; */
/*      for(;i<n_genes;i++) order[i]=n_genes-(i-160);
   for(i=0;i<n_genes;i++) printf("%d %d\n",i,order[i]); */
/*      for(i=0;i<n_genes;i++) order[i]=n_genes-i; */
   jntpeel_sequence (n_genes, order, g_perm, pp);
   free (order);
   free (involved);
}

static int rf_chk (struct Family *fam, struct R_Func *rf, int *order)
{
   int   k1;
   Ind  *kid, *sire, *dam;
   Indlist *id_list;

   sire = fam->sire;
   dam = fam->dam;
   k1 = 0;
   if (rf->id_list[0] == sire->index + 1)
      k1 = 1;
   else if (rf->id_list[0] == dam->index + 1)
      k1 = 1;
   else
   {
      id_list = fam->off;
      while (id_list)
      {
         kid = id_list->current;
         if (order[kid->index] < 2 && rf->id_list[0] == kid->index + 1)
         {
            k1 = 1;
            break;
         }
         id_list = id_list->next;
      }
   }
   return k1;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "Check_RFuncs"
static void Check_RFuncs (struct Family *fam, struct R_Func *rf, int *order,
                          struct Simple_Element *element)
{
   int   k, k1;

   for (k1 = k = 0; k < n_rfuncs[comp - 1]; k++)
      if (!rf[k].flag && rf[k].n_ind <= 2)
         k1 += rf_chk (fam, rf + k, order);
   element->n_rfuncs = k1;
   if (k1)
   {
      if (!(element->index = malloc (sizeof (int) * k1)))
         ABT_FUNC (MMsg);
      for (k1 = k = 0; k < n_rfuncs[comp - 1]; k++)
         if (!rf[k].flag && rf[k].n_ind <= 2 && rf_chk (fam, rf + k, order))
         {
            rf[k].flag = 2;
            element->index[k1++] = k;
         }
   }
}

void locus_free_fam (void)
{
   if (famlist)
      free (famlist);
   if (flist)
      free (flist);
   if (nfam)
      free (nfam);
   famlist = 0;
   flist = 0;
   nfam = 0;
}

void locus_peel_free (void)
{
   int   i, x, x1;
   struct Locus *locus;

   if (!loci)
      return;
   if (singles)
      free (singles);
   x1 = n_markers + (n_tloci ? 1 : 0);
   for (x = 0; x < x1; x++)
   {
      locus = loci + x;
      if (locus->locus_type == MARKER_LOCUS)
      {
         free (locus->allele_trans[0]);
         free (locus->allele_trans);
         free (locus->all_set[0]);
         free (locus->all_set);
      }
      for (comp = 1; comp <= n_compnts; comp++)
      {
         free_peelseq (locus->peel + comp - 1);
         if (locus->r_func[comp - 1])
         {
            for (i = 0; i < locus->n_rfuncs[comp - 1]; i++)
               free (locus->r_func[comp - 1][i].id_list);
            free (locus->r_func[comp - 1]);
         }
      }
      free (locus->peel);
      free (locus->n_rfuncs);
   }
   free (loci[0].r_func);
   free (idflag);
   if (cplist)
      free (cplist);
   if (famlist)
      locus_free_fam ();
   peel_free ();
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "locus_init_fam"
void locus_init_fam (void)
{
   int   i, k, k1, k2, k3, kd;
   Ind  *id, *kid;
   Indlist *id_list;
   Marlist *marlist;

   if (!(idflag = calloc ((size_t) pedsize, sizeof (int))))
      ABT_FUNC (MMsg);
   if (!(nfam = calloc ((size_t) n_compnts, sizeof (int))))
      ABT_FUNC (MMsg);
   /* First step is to count nuclear families in each component */
   for (k = k1 = 0, comp = 1; comp <= n_compnts; comp++)
   {
      for (i = 0, id = nghd; i < pedsize; i++, id++)
         if (id->component == comp && id->marriages)
         {
            kid = id->marriages->current->kids->current;
            if (id == kid->pa)
            {
               marlist = id->marriages;
               while (marlist)
               {
                  nfam[comp - 1]++;
                  marlist = marlist->next;
               }
            }
         }
      k += nfam[comp - 1];
      if (nfam[comp - 1] > k1)
         k1 = nfam[comp - 1];
   }
   if (!(famlist = malloc (sizeof (struct Family) * k)))
      ABT_FUNC (MMsg);
   if (!(flist = malloc (sizeof (void *) * k1)))
      ABT_FUNC (MMsg);
   if (!(cplist = malloc (sizeof (int) * (n_compnts + pedsize))))
      ABT_FUNC (MMsg);
   cp_size = cplist + pedsize;
   /* Now make an array of families for each component in famlist */
   k3 = 0;
   for (i = 0; i < pedsize; i++)
      id_fam[i] = 0;
   for (n_singles = k2 = k = 0, comp = 1; comp <= n_compnts; comp++)
   {
      if (!nfam[comp - 1])
         n_singles++;
      for (i = 0; i < pedsize; i++)
      {
         id = nghd + i;
         if (id->component != comp)
            continue;
         cplist[k2++] = i;
         if (!id->marriages)
            continue;
         kid = id->marriages->current->kids->current;
         if (id == kid->pa)
         {
            marlist = id->marriages;
            while (marlist)
            {
               famlist[k].sire = id;
               k1 = 0;
               id_list = marlist->current->kids;
               famlist[k].dam = id_list->current->ma;
               famlist[k].off = id_list;
               while (id_list)
               {
                  kd = id_list->current->index;
                  id_fam[kd] = famlist + k; /* Keep track of which family an 
                                               individual is an offspring in */
                  k1++;
                  id_list = id_list->next;
               }
               famlist[k++].nkids = k1;
               marlist = marlist->next;
            }
         }
      }
      cp_size[comp - 1] = k2 - k3;
      k3 = k2;
   }
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "locus_peel_init"
void locus_peel_init (int *phase_flag, int quiet_flag)
{
   int   cp, i, id1, j, kd, k, k1, k2, k3, nf, x, x1, m_all, n_all = 0, n_all1,
      all_trans[2], tl_flag;
   int  *all_flag, ch[2], km, kp, *order, *comp_npeel, *y_rec[2];
   Ind  *kid, *pivot, *pivot1, *sire, *dam;
   Indlist *id_list;
   Marlist *marlist;
   struct Locus *locus = 0;
   struct Peelseq_Head *pp, *peelseq_head = 0;
   struct Simple_Element *element;
   struct R_Func *rf, **rfp;
   unsigned long m1, a, b;

   if (!(comp_npeel = calloc ((size_t) n_compnts, sizeof (int))))
      ABT_FUNC (MMsg);
   if (!(order = calloc ((size_t) pedsize, sizeof (int) * 3)))
      ABT_FUNC (MMsg);
   y_rec[0] = order + pedsize;
   y_rec[1] = y_rec[0] + pedsize;
   for (m_all = x = 0; x < n_markers + n_tloci; x++)
   {
      n_all = loci[x].n_all;
      if (n_all > m_all)
         m_all = n_all;
   }
   if (m_all < 2)
      return;
   if (!(gene_obs = malloc (sizeof (int) * pedsize)))
      ABT_FUNC (MMsg);
   if (!(id_fam = malloc (pedsize * sizeof (void *))))
      ABT_FUNC (MMsg);
   if (!(all_flag = malloc (sizeof (int) * m_all)))
      ABT_FUNC (MMsg);
   if (!(temp_set = malloc (sizeof (long) * m_all * 3)))
      ABT_FUNC (MMsg);
   id_set[0] = temp_set + m_all;
   id_set[1] = temp_set + 2 * m_all;
   if (!famlist)
      locus_init_fam ();
   if (!(ngenos = calloc (pedsize, sizeof (int))))
      ABT_FUNC (MMsg);
   if (n_singles)
   {
      if (!quiet_flag)
         printf ("No. singletons: %d\n", n_singles);
      if (!(singles = malloc (sizeof (int) * n_singles)))
         ABT_FUNC (MMsg);
      for (id1 = n_singles = comp = 0; comp < n_compnts; comp++)
      {
         if (!nfam[comp])
         {
            if (cp_size[comp] != 1)
               ABT_FUNC ("Internal error: component with no families\n");
            singles[n_singles++] = comp;
         }
         id1 += cp_size[comp];
      }
   }
   x1 = n_markers + (n_tloci ? 1 : 0);
   if (!(rfp = malloc (sizeof (void *) * x1 * n_compnts)))
      ABT_FUNC (MMsg);                  /* Allocate Rfunction pointers */
   tl_flag = 0;
   for (x = 0; x < n_markers + n_tloci; x++)
   {
      locus = loci + x;
      for (i = 0; i < pedsize; i++)
         gene_obs[i] = 0;
      if (locus->locus_type == MARKER_LOCUS)
      {
         x1 = locus->data_index;
         n_all = locus->n_all;
         locus->sparse_flag = n_all < 3 ? 0 : 1;
         if (!quiet_flag)
            printf ("Processing marker %d (%s stored)\n", x,
                    locus->sparse_flag ? "sparse" : "full");
         if (!(all_set = malloc (sizeof (void *) * n_all)))
            ABT_FUNC (MMsg);
         if (!(all_set[0] = calloc ((size_t) (pedsize * (n_all + 3)),
                                    sizeof (long))))
            ABT_FUNC (MMsg);
         for (i = 1; i < n_all; i++)
            all_set[i] = all_set[i - 1] + pedsize;
         req_set[0] = all_set[i - 1] + pedsize;
         req_set[1] = req_set[0] + pedsize;
         req_set[2] = req_set[1] + pedsize;
         locus->all_set = all_set;
         for (i = 0; i < 3; i++)
            locus->req_set[i] = req_set[i];
         if (!(locus->allele_trans = malloc (sizeof (void *) * n_compnts)))
            ABT_FUNC (MMsg);
         if (!(locus->allele_trans[0] = malloc (sizeof (int) * n_compnts
                                                * n_all)))
            ABT_FUNC (MMsg);
         for (i = 1; i < n_compnts; i++)
            locus->allele_trans[i] = locus->allele_trans[i - 1] + n_all;
      }
      else if (!tl_flag)
      {                                 /* Only do the first trait locus as
                                           peeling sequence is the same */
         tl_flag = 1;
         n_all = locus->n_all;
         req_set[0] = req_set[1] = req_set[2] = 0;
         all_set = 0;
         if (!quiet_flag)
            printf ("Processing trait locus (full stored)\n");
         locus->sparse_flag = 0;
      }
      /* Allele recoding */
      for (k3 = id1 = 0, comp = 1; comp <= n_compnts; comp++)
      {
         if (locus->locus_type == MARKER_LOCUS)
            allele_trans = locus->allele_trans[comp - 1];
         else
            allele_trans = all_trans;
         if (!(peel_trace & TRACE_NO_RECODE) &&
             locus->locus_type == MARKER_LOCUS && !locus->penetrance)
         {                              /* Don't recode for trait loci, or 
                                           markers with penetrance functions                                              */
            /* Find what alleles are used */
            for (i = 0; i < n_all; i++)
               all_flag[i] = 0;
            for (i = 0; i < cp_size[comp - 1]; i++)
            {
               j = cplist[id1 + i];
               if ((j >= 0) && (nghd[j].g_dat))
               {
                  k = nghd[j].g_dat[x1 + x1];
                  if (k)
                     all_flag[k - 1] = 1;
                  k = nghd[j].g_dat[x1 + x1 + 1];
                  if (k)
                     all_flag[k - 1] = 1;
/*
 printf ("\n chkall ind %s loc %d dat  %d %d",
   nghd[j].self, x1, nghd[j].g_dat[x1+x1], nghd[j].g_dat[x1+x1+1]);
 */
               }
            }
            /* Count the no. of alleles (adding 1 for the lumped alleles) */
            for (n_all1 = 1, i = 0; i < n_all; i++)
               n_all1 += all_flag[i];
            if (n_all1 > n_all)
               n_all1 = n_all;          /* Check that we are reducing the 
                                           number of alleles */
         }
         else
            n_all1 = n_all;
         locus->n_alls[comp - 1] = n_all1;
         /* If we are recoding then work out translation for allele codes */
         if (n_all1 < n_all)
         {
            for (j = i = 0; i < n_all; i++)
               if (all_flag[i])
                  allele_trans[j++] = i;
            for (; j < n_all; j++)
               allele_trans[j] = -1;
         }
         else
            for (i = 0; i < n_all; i++)
               allele_trans[i] = i;     /* No translation */
         /* all_flag has (temporarily) the translation table from old to new 
            codes */
         for (i = 0; i < n_all; i++)
            all_flag[i] = n_all1 - i;
         for (i = 0; i < n_all; i++)
         {
            j = allele_trans[i];
            if (j >= 0)
               all_flag[j] = i;
         }
         if ((size_t) n_all1 > LONG_BIT)
            abt (__FILE__, __LINE__,
                 "%s(): No. segregating alleles for marker %d exceeds %d\n",
                 FUNC_NAME, x + 1, LONG_BIT);
         /* Use all_flag to recode haplotypes */
         if (locus->locus_type == MARKER_LOCUS)
         {
            if (n_all1 < n_all)
            {
               for (i = 0; i < cp_size[comp - 1]; i++)
               {
                  j = cplist[id1 + i];
                  if ((j >= 0) && (nghd[j].g_dat))
                  {
                     k = nghd[j].g_dat[x1 + x1];
                     y_rec[0][j] = k ? all_flag[k - 1] + 1 : 0;
                     k = nghd[j].g_dat[x1 + x1 + 1];
                     y_rec[1][j] = k ? all_flag[k - 1] + 1 : 0;
                  }
               }
            }
            else
            {
               for (i = 0; i < cp_size[comp - 1]; i++)
               {
                  j = cplist[id1 + i];
                  if ((j >= 0) && (nghd[j].g_dat))
                  {
                     y_rec[0][j] = nghd[j].g_dat[x1 + x1];
                     y_rec[1][j] = nghd[j].g_dat[x1 + x1 + 1];
                  }
               }
            }
         }
         if (!quiet_flag)
            printf ("Component %d, Size = %d, No. alleles = %d\n", comp,
                    cp_size[comp - 1], n_all1);

/* SCH bug fix; August 2, 1999 ,  correction from SCH made by EAT */
         if (n_all1 < 2)
         {
            id1 += cp_size[comp - 1];
            k3 += nfam[comp - 1];
            continue;
         }

         /* Genotype elimination */
         mask = (1 << n_all1) - 1;
         if (locus->locus_type == MARKER_LOCUS)
         {
            for (i = 0; i < cp_size[comp - 1]; i++)
            {
               k2 = cplist[id1 + i];
               ch[0] = y_rec[0][k2];
               ch[1] = y_rec[1][k2];
               if (ch[0] || ch[1])
                  gene_obs[k2] = 1;
               /* Set all_set mask from genotype data on individual */
               if (ch[0] && ch[1])
               {                        /* both alleles set */
                  for (k = 0; k < n_all1; k++)
                     all_set[k][k2] = 0;
                  all_set[ch[0] - 1][k2] = 1 << (ch[1] - 1);
                  if (phase_flag)
                  {
                     if (!phase_flag[k2])
                        all_set[ch[1] - 1][k2] = 1 << (ch[0] - 1);
                  }
                  else
                     all_set[ch[1] - 1][k2] = 1 << (ch[0] - 1);
               }
               else if (ch[0])
                  all_set[ch[0] - 1][k2] = mask;
               else if (ch[1])
               {
                  m1 = 1 << (ch[1] - 1);
                  for (k1 = 0; k1 < n_all1; k1++)
                     all_set[k1][k2] = m1;
               }
               else
                  for (k1 = 0; k1 < n_all1; k1++)
                     all_set[k1][k2] = mask; /* no alleles set */
            }
         }
         /* Perform iterative genotype elimination, 1 nuclear family at a time */
         nf = nfam[comp - 1];
         for (i = 0; i < nf; i++)
            flist[i] = famlist + (k3++);
         if (locus->locus_type == MARKER_LOCUS && nf)
         {
            for (j = 0; j < nf; j++)
               flist[j]->flag = 1;
            do
            {
               k = 0;
               for (j = 0; j < nf; j++)
               {
                  if (!flist[j]->flag)
                     continue;
                  if (DoFamily (flist[j], n_all1, locus))
                     ABT_FUNC ("Pedigree errors, aborting\n");
                  k = 1;
               }
            }
            while (k);
         }
         /* Fuzzy (set) recoding */
         if (!(peel_trace & TRACE_NO_SETRECODE) &&
             locus->locus_type == MARKER_LOCUS && !locus->penetrance)
         {                              /* Again don't recode for trait 
                                           loci, or markers with penetrance
                                           functions */
            /* Find required allele set for each individual in component (used 
               for set recoding */
            for (i = cp_size[comp - 1] - 1; i >= 0; i--)
            {                           /* Go from offspring -> parents 
                                           though pedigree */
               k2 = cplist[id1 + i];
               req_set[0][k2] = req_set[1][k2] = 0;
               /* Get observed genotypes */
               ch[0] = y_rec[0][k2];
               ch[1] = y_rec[1][k2];
               /* If we have genotypes then we need the observed alleles */
               if (ch[0] || ch[1])
               {
                  for (k = 0; k < 2; k++)
                     if (ch[k])
                     {
                        m1 = 1 << (ch[k] - 1);
                        if (all_set[ch[k] - 1][k2])
                           req_set[X_MAT][k2] |= m1;
                        for (k1 = 0; k1 < n_all1; k1++)
                           if (all_set[k1][k2] & m1)
                              req_set[X_PAT][k2] |= m1;
                     }
               }
               if (ch[0] && ch[1])
                  continue;             /* Skip to next person if this one 
                                           is completely observed */
               /* Loop through all marriages */
               marlist = nghd[k2].marriages;
               while (marlist)
               {
                  /* and kids within each marriage */
                  id_list = marlist->current->kids;
                  while (id_list)
                  {
                     kd = id_list->current->index;
                     if (nghd + k2 == nghd[kd].pa)
                        a = req_set[X_PAT][kd];
                     else
                        a = req_set[X_MAT][kd];
                     for (k1 = 0; k1 < n_all1; k1++)
                     {
                        b = all_set[k1][k2];
                        if (b)
                        {
                           req_set[X_PAT][k2] |= (a & b);
                           m1 = 1 << k1;
                           if (a & m1)
                              req_set[X_MAT][k2] |= m1;
                        }
                     }
                     id_list = id_list->next;
                  }
                  marlist = marlist->next;
               }
            }
            /* Find which individuals we can do set recoding on, lump 
               appropriate alleles together */
            for (i = 0; i < cp_size[comp - 1]; i++)
            {
               k2 = cplist[id1 + i];
               /* Convert from required set to unrequired (but possible) set by
                * xoring with possible set */
               for (b = 0, k1 = 0; k1 < n_all1; k1++)
               {
                  a = all_set[k1][k2];
                  if (!a)
                     continue;
                  m1 = 1 << k1;
                  req_set[X_MAT][k2] ^= m1;
                  b |= a;
               }
               req_set[X_PAT][k2] ^= b;
               /* Find if there are >1 possible but unrequired maternal alleles
                * which can be lumped into allele km */
               k1 = km = 0;
               a = req_set[X_MAT][k2];
               while (a)
               {
                  k1++;
                  if (a & 1)
                     km = k1;
                  a >>= 1;
                  if (km)
                     break;
               }
               if (!a)
               {
                  km = 0;
                  req_set[X_MAT][k2] = 0;
               }
               /* Ditto for paternal alleles */
               k1 = kp = 0;
               a = req_set[X_PAT][k2];
               while (a)
               {
                  k1++;
                  if (a & 1)
                     kp = k1;
                  a >>= 1;
                  if (kp)
                     break;
               }
               if (!a)
               {
                  kp = 0;
                  req_set[X_PAT][k2] = 0;
               }
               /* Lump together possible sets for unused alleles, 
                * zero remainder */
               if (km)
               {                        /* Maternal alleles */
                  for (b = 0, k1 = 0; k1 < n_all1; k1++)
                  {
                     a = all_set[k1][k2];
                     if (!a)
                        continue;
                     m1 = 1 << k1;
                     if (req_set[X_MAT][k2] & m1)
                     {
                        b |= a;
                        all_set[k1][k2] = 0;
                     }
                  }
                  all_set[km - 1][k2] = b;
               }
               if (kp)
               {                        /* Paternal alleles */
                  m1 = 1 << (kp - 1);
                  for (k1 = 0; k1 < n_all1; k1++)
                  {
                     a = (req_set[X_PAT][k2] & all_set[k1][k2]);
                     if (a)
                     {
                        all_set[k1][k2] ^= a;
                        all_set[k1][k2] |= m1;
                     }
                  }
               }
            }
         }
         else if (locus->locus_type == MARKER_LOCUS)
         {                              /* No set recoding */
            for (i = 0; i < cp_size[comp - 1]; i++)
            {
               k2 = cplist[id1 + i];
               req_set[0][k2] = req_set[1][k2] = 0;
            }
         }
         /* Count number of genotypes/individual */
         for (i = 0; i < cp_size[comp - 1]; i++)
         {
            k2 = cplist[id1 + i];
            if (locus->locus_type == MARKER_LOCUS)
            {
               for (k = k1 = 0; k < n_all1; k++)
               {
                  a = all_set[k][k2];
                  while (a)
                  {
                     if (a & 1)
                        k1++;
                     a >>= 1;
                  }
               }
               ngenos[k2] = k1;
            }
            else
               ngenos[k2] = n_all1 * n_all1;
         }
         id1 += cp_size[comp - 1];
      }
      /* Determine peeling sequence */
      if (!(peelseq_head = malloc (sizeof (struct Peelseq_Head) * n_compnts)))
         ABT_FUNC (MMsg);
      if (!(n_rfuncs = calloc ((size_t) n_compnts, sizeof (int))))
         ABT_FUNC (MMsg);
      locus->r_func = r_func = rfp;
      rfp += n_compnts;
      for (i = 0; i < pedsize; i++)
         order[i] = idflag[i] = 0;
      for (k3 = 0, comp = 1; comp <= n_compnts; comp++)
      {
         peelseq_head[comp - 1].type = 0;
         nf = nfam[comp - 1];
         if (!nf)
         {
            r_func[comp - 1] = 0;
            continue;
         }
         pp = peelseq_head + comp - 1;
         for (i = 0; i < nf; i++)
            flist[i] = famlist + (k3++);
         /* Count no. 'connections' to each individual */
         calc_order (nf, flist, order);
         /* (Initial) storage for R-Functions */
         r_func_size = nf + 1;
         if (!(r_func[comp - 1] = rf = malloc (sizeof (struct R_Func)
                                               * r_func_size)))
            ABT_FUNC (MMsg);
         /* Determine a simple peeling sequence where this is possible */
         for (i = 0; i < nf; i++)
         {
            if (peel_trace & TRACE_NO_FAMILYPEEL)
               continue;
            j = 0;
            sire = flist[i]->sire;
            dam = flist[i]->dam;
            pivot = 0;
            if (order[sire->index] > 1)
               pivot = sire;
            if (order[dam->index] > 1)
            {
               if (pivot)
                  j = 1;
               else
                  pivot = dam;
            }
            id_list = flist[i]->off;
            while (!j && id_list)
            {
               kid = id_list->current;
               if (order[kid->index] > 1)
               {
                  if (pivot)
                     j = 1;
                  else
                     pivot = kid;
               }
               id_list = id_list->next;
            }
            /* If !j then a family with only 1 connection to
               the unpeeled portion of the pedigree has been found */
            if (!j)
            {
               /* Set up peeling sequence element */
               if (!(element = malloc (sizeof (struct Simple_Element))))
                  ABT_FUNC (MMsg);
               pp->type = PEEL_SIMPLE;
               pp->ptr.simple = element;
               pp = &element->next;
               pp->type = 0;
               element->sire = sire;
               element->dam = dam;
               element->pivot = pivot;
               element->pivot1 = 0;
               /* Flag R_Functions that are used in operation */
               Check_RFuncs (flist[i], rf, order, element);
               /* A lot of this is only required if complex peeling is required
                */
               if (sire != pivot)
                  idflag[sire->index] |= PEELED;
               if (dam != pivot)
                  idflag[dam->index] |= PEELED;
               if (pivot)
               {
                  element->out_index = n_rfuncs[comp - 1];
                  order[pivot->index]--;
                  idflag[pivot->index] |= WAS_PIVOT;
                  if (n_rfuncs[comp - 1] >= r_func_size)
                  {
                     r_func_size <<= 1;
                     if (!(r_func[comp - 1] = rf =
                           realloc (rf, sizeof (struct R_Func) * r_func_size)))
                        ABT_FUNC (MMsg);
                  }
                  j = n_rfuncs[comp - 1]++;
                  if (!(rf[j].id_list = malloc (sizeof (int) * 2)))
                     ABT_FUNC (MMsg);
                  rf[j].id_list[0] = pivot->index + 1;
                  rf[j].id_list[1] = -1 - pivot->index;
                  rf[j].n_ind = 2;
                  rf[j].flag = 0;
               }
               else
                  element->out_index = -1;
               id_list = flist[i]->off;
               if (!(element->off = malloc (sizeof (Indlist)
                                            * flist[i]->nkids)))
                  ABT_FUNC (MMsg);
               if (flist[i]->nkids > max_peel_off)
                  max_peel_off = flist[i]->nkids;
               j = 0;
               while (id_list)
               {
                  kid = id_list->current;
                  id_list = id_list->next;
                  element->off[j++].current = kid;
                  element->off[j - 1].next = element->off + j;
                  if (kid != pivot)
                     idflag[kid->index] |= PEELED;
               }
               element->off[j - 1].next = 0;
               /* Look for another family to peel */
               flist[i] = flist[--nf];
               i = -1;
            }
         }
         /* Some families left - this means there are loops */
         /* Try to partially peel some families */
         for (i = 0; i < nf; i++)
         {
            if (peel_trace & (TRACE_NO_FAMILYPEEL | TRACE_NO_PARTIALPEEL))
               continue;
            sire = flist[i]->sire;
            dam = flist[i]->dam;
            j = 0;
            pivot = pivot1 = 0;
            if (order[dam->index] > 1)
               pivot = dam;
            if (order[sire->index] > 1)
            {
               if (pivot)
                  pivot1 = sire;
               else
                  pivot = sire;
            }
            k1 = 0;
            id_list = flist[i]->off;
            while (id_list)
            {
               kid = id_list->current;
               if (order[kid->index] > 1)
               {
                  if (!pivot)
                     pivot = kid;
                  else if (!pivot1)
                     pivot1 = kid;
                  else
                     j = 1;
               }
               id_list = id_list->next;
            }
            if (j)
            {                           /* More than 2 potential pivots */
               id_list = flist[i]->off;
               while (id_list)
               {
                  kid = id_list->current;
                  if (order[kid->index] < 2)
                  {
                     k1++;
                     if (j)
                     {
                        j = 0;
                        pivot = pivot1 = 0;
                        if (order[dam->index] > 1)
                           pivot = dam;
                        if (order[sire->index] > 1)
                        {
                           if (pivot)
                              pivot1 = sire;
                           else
                              pivot = sire;
                        }
                     }
                  }
                  id_list = id_list->next;
               }
            }
            if (!j && pivot)
            {
               if (!(element = malloc (sizeof (struct Simple_Element))))
                  ABT_FUNC (MMsg);
               pp->type = PEEL_SIMPLE;
               pp->ptr.simple = element;
               pp = &element->next;
               pp->type = 0;
               element->sire = sire;
               element->dam = dam;
               element->pivot = pivot;
               element->pivot1 = pivot1;
               /* Flag R_Functions that are used in operation */
               Check_RFuncs (flist[i], rf, order, element);
               element->out_index = n_rfuncs[comp - 1];
               if (n_rfuncs[comp - 1] >= r_func_size)
               {
                  r_func_size <<= 1;
                  if (!(r_func[comp - 1] = rf =
                        realloc (rf, sizeof (struct R_Func) * r_func_size)))
                     ABT_FUNC (MMsg);
               }
               k = n_rfuncs[comp - 1]++;
               if (!(rf[k].id_list = malloc (sizeof (int) * 4)))
                  ABT_FUNC (MMsg);
               rf[k].id_list[0] = pivot->index + 1;
               rf[k].id_list[1] = -1 - pivot->index;
               idflag[pivot->index] |= WAS_PIVOT;
               if (pivot1)
               {
                  rf[k].id_list[2] = pivot1->index + 1;
                  rf[k].id_list[3] = -1 - pivot1->index;
                  rf[k].n_ind = 4;
                  idflag[pivot1->index] |= WAS_PIVOT;
               }
               else
                  rf[k].n_ind = 2;
               rf[k].flag = 0;
               id_list = flist[i]->off;
               j = 0;
               if (flist[i]->nkids > max_peel_off)
                  max_peel_off = flist[i]->nkids;
               if (k1)
               {                        /* We're not peeling all kids */
                  if (!(element->off = malloc (sizeof (Indlist) * k1)))
                     ABT_FUNC (MMsg);
                  while (id_list)
                  {
                     kid = id_list->current;
                     id_list = id_list->next;
                     if (order[kid->index] > 1)
                        continue;
                     element->off[j++].current = kid;
                     element->off[j - 1].next = element->off + j;
                     idflag[kid->index] |= PEELED;
                  }
                  element->off[j - 1].next = 0;
               }
               else
               {
                  if (!(element->off = malloc (sizeof (Indlist) *
                                               flist[i]->nkids)))
                     ABT_FUNC (MMsg);
                  while (id_list)
                  {
                     kid = id_list->current;
                     id_list = id_list->next;
                     element->off[j++].current = kid;
                     element->off[j - 1].next = element->off + j;
                     if (kid != pivot && kid != pivot1)
                        idflag[kid->index] |= PEELED;
                  }
                  element->off[j - 1].next = 0;
               }
               if (dam != pivot)
                  idflag[dam->index] |= PEELED;
               if (sire != pivot && sire != pivot1)
                  idflag[sire->index] |= PEELED;
               if (!k1)
                  flist[i--] = flist[--nf];
            }
         }
         for (i = 0; i < n_rfuncs[comp - 1]; i++)
            if (!rf[i].flag)
               break;
         if (nf || i < n_rfuncs[comp - 1])
            find_sequence (nf, flist, pp, rf);
         pp = peelseq_head + comp - 1;
         if (peel_trace & TRACE_DUMP_SEQUENCE)
            print_sequence (stdout, pp, r_func[comp - 1]);
         cp = 0;
         while (pp->type)
         {
            cp++;
            pp = &pp->ptr.simple->next;
         }
         if (cp > comp_npeel[comp - 1])
            comp_npeel[comp - 1] = cp;
      }
      locus->peel = peelseq_head;
      locus->n_rfuncs = n_rfuncs;
   }
   if (n_tloci > 1)
   {
      for (j = i = 0; i < n_markers + n_tloci; i++)
         if (loci[i].locus_type == TRAIT_LOCUS)
         {
            if (j++)
               locus->peel = peelseq_head;
            else
               peelseq_head = locus->peel;
         }
   }
   free (id_fam);
   free (ngenos);
   free (order);
   free (all_flag);
   free (temp_set);
   free (gene_obs);
   locus_peel_alloc (loci, n_markers + n_tloci, comp_npeel);
   free (comp_npeel);

   return;
}
