/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
#include "lsample.h"
#include "msample.h"

/**
 * Get initial xx0, xx1 using locus-by-locus sampling or sequential imputation.
 * Then, from these, set the initial FGLs (founder genome labels)
 *
 * As for sampling(), the mm-flag indicates whether locus 0 (mm=0) should
 *    be set/sampled  or not (mm=1) : EAT  July 2005.
 */
void initInherit (int mm)
{
   int   i, j;

   prntStatusSMC ();

   set_seg_mode (0);

   peel_trace = 0;

   /* set up initial maternal and inheritance segregation indicators 
    * use locus-by-locus sampling or sequential imputation 
    *                                            xx0[][], xx1[][]  */
   if (SmplSetup == INDEP_LOC)
   {
      set_lsmpl (0, mm, NMarks);
      prntStatusIMC ("locus-by-locus sampling");
   }
   else if (SmplSetup == SEQIMP)
   {
      if (TraitTyp != QUANTITATIVE)
         do_seq_imp (NSeqImp, mm, PedSize, NMarks + 1, 1.0);
      else
         do_seq_imp (NSeqImp, mm, PedSize, NMarks + 1, loci[0].ve);
      prntStatusIMC ("sequential imputation");
   }

   peel_trace = 0;
   if (peel_trace)
   {
      printf ("\n check initial inheritance assignments \n");
      fflush (stdout);
      for (i = 0; i < PedSize; ++i)
         if (!(isNullNode (nghd[i].pa)))
         {
            printf (" %s", nghd[i].self);
            fflush (stdout);
            for (j = mm; j < NMarks + 1; ++j)
               printf ("  %d %d", xx1[j][i], xx0[j][i]);
            printf ("\n \n");
            fflush (stdout);
         }
   }

   /*set the initial FGLs
    *                                            gg0[][], gg1[][]             */
   for (j = mm; j < NMarks + 1; ++j)
      set_FGL (j);

   if (peel_trace)
   {
      printf ("\n check initial fgl assignments \n");
      fflush (stdout);
      for (i = 0; i < PedSize; ++i)
      {
         printf (" %s", nghd[i].self);
         fflush (stdout);
         for (j = mm; j < NMarks + 1; ++j)
            printf ("  %d %d", gg1[j][i], gg0[j][i]);
         printf ("\n \n");
         fflush (stdout);
      }
   }

   peel_trace = 0;

   return;
}

/**
 * set up founder genome label peeling.
 */
void setFGLpeel (int mm)
{
   int   j, k;

   k = 0;
   for (j = mm; j < NMarks + 1; ++j)
   {
      /* Allocate memory for gen_pen() */
      if (loci[j].locus_type == TRAIT_LOCUS)
         alloc_gen_pen (nFGL, j);
      else if (!k++)
         alloc_seg_pen (nFGL, NMarks + 1);
   }
   return;
}

/**
 * set up prob[][] array for storing seg_pen results.
 */
double **initPenProb (int mm)
{
   double **p;
   int   c, j, err;

   /* allocate prob[][] array for storing seg_pen results */
   ald_nc_nl (&p);

   for (j = mm; j <= NMarks; ++j)
      for (c = 0; c < nCompnt; ++c)
         p[c][j] = seg_pen2 (c, Qq[j], j, nFGL, &err);

   if (DebugMain)
      dbgProb (p);

   return p;
}
